/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <cmath>
#include <iostream>
#include "Debug.h"
#include <cassert>

template<typename T> class Vec3D;

/** OPERATORS */

template <class T> bool operator!= (const Vec3D<T> & p1, const Vec3D<T> & p2) {
    return (p1[0] != p2[0] || p1[1] != p2[1] || p1[2] != p2[2]);
}

template <class T> const Vec3D<T> operator* (const Vec3D<T> & p, float factor) {
    return Vec3D<T> (p[0] * factor, p[1] * factor, p[2] * factor);
}

template <class T> const Vec3D<T> operator* (float factor, const Vec3D<T> & p) {
    return Vec3D<T> (p[0] * factor, p[1] * factor, p[2] * factor);
}

template <class T> const Vec3D<T> operator* (const Vec3D<T> & p1, const Vec3D<T> & p2) {
    return Vec3D<T> (p1[0] * p2[0], p1[1] * p2[1], p1[2] * p2[2]);
}

template <class T> const Vec3D<T> operator+ (const Vec3D<T> & p1, const Vec3D<T> & p2) {
    return Vec3D<T> (p1[0] + p2[0], p1[1] + p2[1], p1[2] + p2[2]);
}

template <class T> const Vec3D<T> operator- (const Vec3D<T> & p1, const Vec3D<T> & p2) {
    return Vec3D<T> (p1[0] - p2[0], p1[1] - p2[1], p1[2] - p2[2]);
}

template <class T> const Vec3D<T> operator- (const Vec3D<T> & p) {
    return Vec3D<T> (-p[0], -p[1], -p[2]);
}

template <class T> const Vec3D<T> operator/ (const Vec3D<T> & p, float divisor) {
    return Vec3D<T> (p[0]/divisor, p[1]/divisor, p[2]/divisor);
}

template <class T> bool operator== (const Vec3D<T> & p1, const Vec3D<T> & p2) {
    return (p1[0] == p2[0] && p1[1] == p2[1] && p1[2] == p2[2]);
}

template <class T> bool operator< (const Vec3D<T> & a, const Vec3D<T> & b) {
    return (a[0] < b[0] && a[1] < b[1] && a[2] < b[2]);
}

template <class T> bool operator>= (const Vec3D<T> & a, const Vec3D<T> & b) {
    return (a[0] >= b[0] || a[1] >= b[1] || a[2] >= b[2]);
}


/**
 * Vector in 3 dimensions, with basics operators overloaded.
 */
template <typename T>
/**
 * @brief The Vec3D class
 */
class Vec3D {
public:
    inline Vec3D (void)	{
        p[0] = p[1] = p[2] = T ();
    }
    inline Vec3D (T p0, T p1, T p2) {
        p[0] = p0;
        p[1] = p1;
        p[2] = p2;
    }

    inline Vec3D (const Vec3D & v) {
        init (v[0], v[1], v[2]);
    }

    inline Vec3D (T* pp) {
        p[0] = pp[0];
        p[1] = pp[1];
        p[2] = pp[2];
    }
    inline Vec3D (const T* pp) {
        p[0] = pp[0];
        p[1] = pp[1];
        p[2] = pp[2];
    }

    // ---------
    // Operators
    // ---------
    inline T& operator[] (int Index) {
        return (p[Index]);
    }

    inline const T& operator[] (int Index) const {
        return (p[Index]);
    }

    inline Vec3D& operator= (const Vec3D & P) {
        p[0] = P[0];
        p[1] = P[1];
        p[2] = P[2];
        return (*this);
    }
    inline Vec3D& operator+= (const Vec3D & P) {
        p[0] += P[0];
        p[1] += P[1];
        p[2] += P[2];
        return (*this);
    }

    inline Vec3D& operator-= (const Vec3D & P) {
        p[0] -= P[0];
        p[1] -= P[1];
        p[2] -= P[2];
        return (*this);
    }

    inline Vec3D& operator*= (const Vec3D & P) {
        p[0] *= P[0];
        p[1] *= P[1];
        p[2] *= P[2];
        return (*this);
    }

    inline Vec3D& operator*= (T s) {
        p[0] *= s;
        p[1] *= s;
        p[2] *= s;
        return (*this);
    }

    inline Vec3D& operator/= (const Vec3D & P) {
        p[0] /= P[0];
        p[1] /= P[1];
        p[2] /= P[2];
        return (*this);
    }

    inline Vec3D& operator/= (T s) {
        p[0] /= s;
        p[1] /= s;
        p[2] /= s;
        return (*this);
    }


    //---------------------------------------------------------------

    inline Vec3D & init (T x, T y, T z) {
        p[0] = x;
        p[1] = y;
        p[2] = z;
        return (*this);
    }

    /**
     * @brief getSquaredLength
     * @return the vector squared length
     */
    inline T getSquaredLength() const {
        return (dotProduct (*this, *this));
    }

    /**
     * @brief getLength
     * @return the Vec3D length
     */
    inline T getLength() const {
        return (T)sqrt (getSquaredLength());
    }

    /**
     * @brief normalize
     * @return the length after normalization
     */
    inline T normalize (void) {
        T length = getLength();
        if (length == 0.0f)
            return 0;
        T rezLength = 1.0f / length;
        p[0] *= rezLength;
        p[1] *= rezLength;
        p[2] *= rezLength;
        return length;
    }

    /**
     * @brief fromTo set the vector to go from p1 to p2
     * @param P1
     * @param P2
     */
    inline void fromTo (const Vec3D & P1, const Vec3D & P2) {
        p[0] = P2[0] - P1[0];
        p[1] = P2[1] - P1[1];
        p[2] = P2[2] - P1[2];
    }

    /**
     * @brief transProduct
     * @param v
     * @return the dotProduct between v and this
     */
    inline float transProduct (const Vec3D & v) const {
        return (p[0]*v[0] + p[1]*v[1] + p[2]*v[2]);
    }

    /**
     * @brief dotProduct
     * @param a
     * @param b
     * @return the dotProduct between a and b
     */
    static inline T dotProduct(const Vec3D & a, const Vec3D & b) {
        return (a[0] * b[0] + a[1] * b[1] + a[2] * b[2]);
    }

    /**
     * @brief getTwoOrthogonals set u and v so that they are orthogonals to this
     * @param u
     * @param v
     */
    inline void getTwoOrthogonals (Vec3D & u, Vec3D & v) const {
        if (fabs(p[0]) < fabs(p[1])) {
            if (fabs(p[0]) < fabs(p[2]))
                u = Vec3D (0, -p[2], p[1]);
            else
                u = Vec3D (-p[1], p[0], 0);
        } else {
            if (fabs(p[1]) < fabs(p[2]))
                u = Vec3D (p[2], 0, -p[0]);
            else
                u = Vec3D(-p[1], p[0], 0);
        }
        v = crossProduct (*this, u);
    }

    /**
     * @brief segment
     * @param a
     * @param b
     * @return same as fromTo
     */
    static inline Vec3D segment (const Vec3D & a, const Vec3D & b) {
        Vec3D r;
        r[0] = b[0] - a[0];
        r[1] = b[1] - a[1];
        r[2] = b[2] - a[2];
        return r;
    }

    /**
     * @brief crossProduct
     * @param a
     * @param b
     * @return the cross product between a and b
     */
    static inline Vec3D crossProduct(const Vec3D & a, const Vec3D & b) {
        Vec3D result;
        result[0] = a[1] * b[2] - a[2] * b[1];
        result[1] = a[2] * b[0] - a[0] * b[2];
        result[2] = a[0] * b[1] - a[1] * b[0];
        return(result);
    }

    /**
     * @brief squaredDistance
     * @param v1
     * @param v2
     * @return the squared distance between v1 and v2
     */
    static inline T squaredDistance (const Vec3D &v1, const Vec3D &v2) {
        Vec3D tmp = v1 - v2;
        return (tmp.getSquaredLength());
    }
    /**
     * @brief distance
     * @param v1
     * @param v2
     * @return the distance between v1 and v2
     */
    static inline T distance (const Vec3D &v1, const Vec3D &v2) {
        Vec3D tmp = v1 - v2;
        return (tmp.getLength());
    }

    /**
     * @brief interpolate
     * @param u
     * @param v
     * @param alpha
     * @return a vector that is at (1-a)u+av
     */
    static inline Vec3D interpolate (const Vec3D & u, const Vec3D & v, T alpha) {
        return (u * (1.0f - alpha) + v * alpha);
    }

    /**
     * @brief cartesianToPolar
     * @param v
     * @return [0] = length, [1] = angle with z-axis, [2] = angle of projection into x,y, plane with x-axis
     */
    static inline Vec3D cartesianToPolar (const Vec3D &v) {
        Vec3D polar;
        polar[0] = v.getLength();
        if (v[2] > 0.0f)
            polar[1] = (T) atan (sqrt (v[0] * v[0] + v[1] * v[1]) / v[2]);
        else if (v[2] < 0.0f)
            polar[1] = (T) atan (sqrt (v[0] * v[0] + v[1] * v[1]) / v[2]) + M_PI;
        else
            polar[1] = M_PI * 0.5f;
        if (v[0] > 0.0f)
            polar[2] = (T) atan (v[1] / v[0]);
        else if (v[0] < 0.0f)
            polar[2] = (T) atan (v[1] / v[0]) + M_PI;
        else if (v[1] > 0)
            polar[2] = M_PI * 0.5f;
        else
            polar[2] = -M_PI * 0.5;
        return polar;
    }

    /**
     * @brief polarToCartesian
     * @param v
     * v[0] = length
     * v[1] = angle with z-axis
     * v[2] = angle of projection into x,y, plane with x-axis
     * @return
     */
    static inline Vec3D polarToCartesian (const Vec3D & v) {
        Vec3D cart;
        cart[0] = v[0] * (T) sin (v[1]) * (T) cos (v[2]);
        cart[1] = v[0] * (T) sin (v[1]) * (T) sin (v[2]);
        cart[2] = v[0] * (T) cos (v[1]);
        return cart;
    }

    /**
     * @brief projectOn
     * @param N
     * @param P
     * @return the projection of this onto the line that go through P and of normal vector N
     */
    inline Vec3D projectOn (const Vec3D & N, const Vec3D & P) const {
        T w = dotProduct (((*this) - P), N);
        return (*this) - (N * w);
    }

    /**
     * @brief projectOntoVector
     * @param v1
     * @param v2
     * @return the projection of v1 on v2
     */
    static inline Vec3D projectOntoVector (const Vec3D & v1, const Vec3D & v2) {
        return v2 * dotProduct (v1, v2);
    }

    /**
     * @brief transformIn
     * @param pos
     * @param n
     * @param u
     * @param v
     * @return express this in the frame (pos, u, v, n)
     */
    inline Vec3D transformIn (const Vec3D & pos, const Vec3D & n, const Vec3D & u, const Vec3D & v) const {
        Vec3D q = (*this) - pos;
        return Vec3D (u[0]*q[0] + u[1]*q[1] + u[2]*q[2],
                      v[0]*q[0] + v[1]*q[1] + v[2]*q[2],
                      n[0]*q[0] + n[1]*q[1] + n[2]*q[2]);
    }

    /**
     * @brief applyRotation rotate p around axis with a teta angle (in radians)
     * @param axis
     * @param teta
     */
    inline void applyRotation(Vec3D axis, T teta)
    {
        axis.normalize();
        float R[9];
        R[0]=axis[0]*axis[0]*(1-cos(teta))+cos(teta);           R[1]=axis[1]*axis[0]*(1-cos(teta))-axis[2]*sin(teta);    R[2]=axis[2]*axis[0]*(1-cos(teta))+ axis[1]*sin(teta);
        R[3]=axis[0]*axis[1]*(1-cos(teta))+axis[2]*sin(teta);   R[4]=axis[1]*axis[1]*(1-cos(teta))+cos(teta);            R[5]=axis[2]*axis[1]*(1-cos(teta))- axis[0]*sin(teta);
        R[6]=axis[0]*axis[2]*(1-cos(teta))-axis[1]*sin(teta);   R[7]=axis[1]*axis[2]*(1-cos(teta))+axis[0]*sin(teta);    R[8]=axis[2]*axis[2]*(1-cos(teta))+ cos(teta);

        Vec3D res;
        res[0] = R[0]*p[0] +  R[1]*p[1] +  R[2]*p[2];
        res[1] = R[3]*p[0] +  R[4]*p[1] +  R[5]*p[2];
        res[2] = R[6]*p[0] +  R[7]*p[1] +  R[8]*p[2];

        (*this) = res;
    }

    /**
     * @brief applyTranslation translate this from dX
     * @param dX
     */
    inline void applyTranslation( Vec3D dX)
    {
        (*this) += dX;
    }

    /**
     * @brief applyScale
     * @param _scale
     */
    inline void applyScale(Vec3D _scale)
    {
        (*this) *=_scale;
    }

    /**
     * @brief expressInBasis
     * @param u
     * @param v
     * @param w
     * @return p expressed in (u,v,w) frame
     */
    inline Vec3D expressInBasis(Vec3D u, Vec3D v, Vec3D w)
    {
        return( Vec3D(dotProduct(u, *this),
                      dotProduct(v, *this),
                      dotProduct(w, *this)));
    }

    /**
     * @brief projectionOnPlane
     * @param p1
     * @param p2
     * @param p3
     * @param distToPlane
     * @return
     */
    inline Vec3D projectionOnPlane(Vec3D p1, Vec3D p2, Vec3D p3, float & distToPlane)
    {
        Vec3D n = crossProduct(p2-p1, p3-p1);
        n.normalize();
        float d = -n[0]*p1[0]-n[1]*p1[1]-n[2]*p1[2];
        float k = (-d-dotProduct((*this),n))/dotProduct(n,n);

        // h = dotProduct((*this)-p1,n)*n + (*this);
        distToPlane = (k*n).getLength();
        return ((*this)+k*n);
    }

    /**
     * @brief pointProjectedOnTriangle
     * @param p1 triangle vertex
     * @param p2 triangle vertex
     * @param p3 triangle vertex
     * @param d0 (out) barycentric coordinates of the projection
     * @param d1 (out) barycentric coordinates of the projection
     * @param d2 (out) barycentric coordinates of the projection
     * @param distToPlane (out) distance between the point and it projection
     * @return true if the projection of the point is on the triangle
     */
    inline bool pointProjectedOnTriangle(Vec3D p1, Vec3D p2, Vec3D p3,
                                         double& d0, double& d1, double& d2, float& distToPlane)
    {

        Vec3D h = projectionOnPlane(p1, p2, p3, distToPlane);
        T u, v;
        Vec3D v0 = p3-p1;
        Vec3D v1 = p2-p1;
        Vec3D v2 = h-p1;

        u = (dotProduct(v1,v1)*dotProduct(v2,v0)-dotProduct(v1,v0)*dotProduct(v2,v1)) /
                (dotProduct(v0,v0)*dotProduct(v1,v1) - dotProduct(v0,v1)*dotProduct(v1,v0));
        v = (dotProduct(v0,v0)*dotProduct(v2,v1)-dotProduct(v0,v1)*dotProduct(v2,v0)) /
                (dotProduct(v0,v0)*dotProduct(v1,v1) - dotProduct(v0,v1)*dotProduct(v1,v0));
        d2 = u;
        d1 = v;
        d0 = 1 -u -v;
        if ( (u>= (T) 0.) && (v>=(T) 0.) && (u+v<=(T) 1.) )
            return true;
        else
            return false;
    }

protected:
    T p[3];
};

template <class T>  inline void swap (Vec3D<T> & P, Vec3D<T> & Q) {
    Vec3D<T> tmp = P;
    P = Q;
    Q = tmp;
}

template <class T> std::ostream & operator<< (std::ostream & output, const Vec3D<T> & v) {
    output << v[0] << " " << v[1] << " " << v[2];
    return output;
}

template <class T> std::istream & operator>> (std::istream & input, Vec3D<T> & v) {
    input >> v[0] >> v[1] >> v[2];
    return input;
}

typedef Vec3D<float> Vec3Df;
typedef Vec3D<double> Vec3Dd;
typedef Vec3D<int> Vec3Di;

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
