/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SELECTION_H
#define SELECTION_H
#include "Mesh.h"
#include "UserStroke.h"
#include "MathTools.h"
#include "CotanLaplacian.h"

/**
 * @class Selection is a virtual class to represent a selection
 */

#define SPY_TIME true

enum SIM_HIST_DIST{
    L1 = 0,
    L2 = 1,
    CHI2 = 2,
    CHI2N = 3
};

class Selection : public QObject, public Mesh
{
    Q_OBJECT
public:
    Selection():Mesh(){}

    virtual ~Selection(){
    }


    /**
     * @fn init
     * @param m the mesh
     * @brief initialize all the necessary values from the mesh m
     */
    void init(const Mesh & m){
        vertices = m.getVertices();
        triangles = m.getTriangles();
        oneRing = m.getOneRing();
        oneRingTr = m.getOneRingTr();
        faceNeighbors = m.getFaceNeighbors();
        average_edge_halfsize = m.getAverageEdgeHalfSize();

        connectedCompRepresentatives = m.getConnectedCompRepresentatives();
        vertexConnectedComp =  m.getVerticeConnectedCompInd();
        vertexIsOnShapeBoundary = m.getVertexIsOnShapeBoundary();
        faceInSelectionK.clear();
        for (unsigned int i = 0; i<triangles.size(); i++)
            faceInSelectionK.push_back(-1.f);
        histoDist = L2;
        nb_of_local_sel = 0;
        nb_of_global_sel = 0;
    }

    /**
     * @fn startSelection
     * @param stroke the first user stroke
     * @brief start a new selection from the user stroke
    */
    virtual void startSelection(UserStroke & stroke) = 0;

    /**
     * @fn update selection
     * @param stroke the user stroke
     * @brief update the current selection based on the user stroke
    */
    virtual void updateSelection(UserStroke & stroke) = 0;

    /**
     * @fn endSelection
     * @param stroke the first user stroke
     * @brief update the current selection from the user stroke and stop it.
    */
    virtual void endSelection(UserStroke & stroke) = 0;

    /**
     * @fn eraseSelection
     * @param stroke the user stroke
     * @brief erase a part of the selection based on the user stroke
    */
    virtual void eraseSelection(UserStroke & stroke) = 0;

    /**
     * @fn refineSelection
     * @param stroke the user stroke
     * @brief refine the selection boundary from the user stroke
    */
    virtual void refineSelection(UserStroke & stroke) = 0;

    /**
     * @fn findSimilarSelection
     * @param stroke the user stroke
     * @brief find the k most similar selections
     */
    virtual void findSimilarSelections(UserStroke& _stroke) = 0;

    /**
     * @fn findSimilarSelection
     * @param stroke the user stroke
     * @brief find the k most similar selections
     */
    virtual void changeSimiliarityThreshold(int val) = 0;

    /**
     * @fn setSimilarityThreshold
     * @param val the user defined threshold
     */
    void setSimiliarityThreshold(int val){
        nb_of_similar_sel = val;
    }

    /**
     * @fn getSelections
     * @return the selections
     */
    const std::vector<float> & getSelections(){return faceInSelectionK;}

    /**
     * @fn setSelections
     * @brief set the selection
     */
    void setSelections(const std::vector<float> & sel){
        assert (sel.size() == faceInSelectionK.size());
        for(unsigned int i = 0; i < faceInSelectionK.size(); i++)
            faceInSelectionK[i] = sel[i];
    }

    virtual void clearSelections()=0;
    virtual void clear()=0;

    /**
     * @fn getSimSelectionsNb
     * @return the number of similar selections
     */
    unsigned int getSimSelectionsNb(){return simSelectionsNb;}

    /**
     * @fn changeHistSimDist
     * @param val type of distance used
     * @brief change the distance used to compare histograms (L1, L2, CHI2..)
     */
    void changeHistSimDist(int val){
        Trace(1, " Selection::changeHistSimDist "<< val);
        histoDist = (SIM_HIST_DIST) val;
        recomputeSimilarityDist();
    }

    /**
     * @fn getRefHistogram
     * @return the histogram (descriptor) of the reference vertex (center of the reference selection)
     */
    virtual QPixmap getRefHistogram() = 0;

    /**
     * @fn getUnderMouseHistogram
     * @return the histogram (descriptor) of the vertex under the mouse
     */
    virtual QPixmap getUnderMouseHistogram(unsigned int v, float & dist) = 0;

    /**
     * @fn setMeshLaplacian
     */
    void setMeshLaplacian(CotanLaplacian* l){
        _meshLaplacian = l;
    }

    virtual void saveTimings(const QString &filename) = 0;
protected :
    virtual void recomputeSimilarityDist() = 0;

    std::vector<float> faceInSelectionK; /**< each face that is part of a selection has a faceInselectionK > 0 */
    unsigned int simSelectionsNb; /**< number of similar selections */
//    float similarity_threshold;
    SIM_HIST_DIST histoDist; /**< type of distance used to compare the descriptors */

    unsigned int numberOfStrokePnts; /**< number of stroke points */
    unsigned int numberOfBackGroundPnts; /**< number of eraser stroke points */
    unsigned int numberOfRefineStrokePnts; /**< number of refine stroke points */

    CotanLaplacian* _meshLaplacian; /**< mesh laplacian */

    std::vector<float> timing_precomputation; /**< precomputation timing */
    std::vector<std::pair< unsigned int,float > > timing_local; /**< selection timing */
    std::vector<std::pair< unsigned int, float> > timing_global; /**< expansion timing */

    unsigned int nb_of_local_sel; /**< number of selections performed by the user */
    unsigned int nb_of_global_sel; /**< total number of selection (user defined +automatically detected) */
    unsigned int nb_of_similar_sel; /**< number of similar selections required by the user */
signals:
    void similarityMapComputed(int i);
};

#endif // SELECTION_H
