/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "PatchSelection.h"
#include "Distance.h"
#include <QFile>

// number of bins in the PSC descriptor
#define DIST_BIN_NB 10
#define ANGLE_BIN_NB 10

// similarity treshold
#define SIMILARITY_THRESHOLD 0.4

// useful for debug mode
#define SEE_SVM false
#define SEE_SEEDS false

using namespace std;


bool floatCompare (pair< float, pair<int, unsigned int> > i,
                   pair< float, pair<int, unsigned int> > j)
{ return (i.first >j.first ); }

void PatchSelection::init(const Mesh &m){
    Trace(2, "PatchSelection::init ");
    Selection::init(m);
    LcrossDistUsed = false;
    // shape context initialization
    Hist2Df sc = Hist2Df(DIST_BIN_NB, ANGLE_BIN_NB, 0, 1, 0, M_PI);
    for (unsigned int i = 0; i < vertices.size(); i++)
    {
        _shapeContexts.push_back(sc);
        similarity_dist.push_back(-1.f);
        _sc_computed.push_back(false);
    }
    computeBoundingBox();
    numberOfStrokePnts = 0;
    numberOfBackGroundPnts = 0;
    numberOfRefineStrokePnts = 0;
    nb_of_similar_sel = 1;
    angle = 1.f;
    trianglesUnderMouse.clear();
}

void PatchSelection::clear(){
    Trace(2, "PatchSelection::clear ");
    Mesh::clear();
    _shapeContexts.clear();
    similarity_dist.clear();
    _sc_computed.clear();
    refCenter = Vec3Df(FLT_MAX, FLT_MAX, FLT_MAX);
    refVertex = INT_MAX;
    ind_vertex = INT_MAX;
    ICPpointsRef.clear();
    ICPpointsRefTransformed.clear();
    trianglesUnderMouse.clear();
    timing_precomputation.clear();
    timing_global.clear();
    timing_local.clear();
}


void PatchSelection::saveTimings(const QString &filename){
    Trace(2, "PatchSelection::saveTimings ");
    std::string file_n = qPrintable(filename);
    file_n.append("_patch");
    ofstream out (file_n.c_str ());

    out << "V: " << vertices.size() << " T: "<< triangles.size() << "\n";
    out << "------------------------------------- \n";
    out << "-------     local    ---------------- \n";
    out << "------------------------------------- \n";
    int previous_sel = -1;
    float avg = 0;
    for (unsigned int i = 0; i < timing_local.size(); i ++)
    {
        if ((int) timing_local[i].first != previous_sel){
            out <<"\n" << timing_local[i].first<< " - ";
            previous_sel = timing_local[i].first;
        }
        out << (float) timing_local[i].second << " ";
        avg +=  (float) timing_local[i].second ;
    }
    out << "\n" << " AVERAGE " << avg/timing_local.size();
    out << "\n";
    out << "------------------------------------- \n";
    out << "-------    global    ---------------- \n";
    out << "------------------------------------- \n";

    for (unsigned int i = 0; i < timing_global.size(); i++)
    {
            out <<"\n" << timing_global[i].first<< " - " << (float) timing_global[i].second << " ";
    }

    out << "\n";
    out << "------------------------------------- \n";
    out << "-------   global detailled----------- \n";
    out << "------------------------------------- \n";

    int previous_sim_sel = -1;
    unsigned int j = 0;
    for (unsigned int i = 0; i < timing_SC.size(); i ++)
    {
        out << timing_SC[i].second.sel_nb << " - diskRadius " << timing_SC[i].first << " timing SC "
            << timing_SC[i].second.time <<  " timing SVM train " << timing_SVM_train[i].time<< "\n";
        previous_sim_sel = timing_SC[i].second.sel_nb ;
        while ( (int) timing_SVM_predict[j].sel_nb == previous_sim_sel)
        {
            out << "     " << timing_SVM_predict[j].sim_nb <<" : SVM predict :" << timing_SVM_predict[j].time
                << " ICP " << timing_ICP[j].time << "\n";
            j++;
        }
        out << "\n";
    }
    out.close();
}


//***************************************************
//                  SELECTION
//***************************************************

void PatchSelection::startSelection(UserStroke & stroke)
{
    Trace(3, "PatchSelection::startSelection ");
    QTime timer;
    if (SPY_TIME)
        timer.start();

    for (unsigned int i = 0; i < triangles.size(); i++)
        trianglesUnderMouse.push_back(-1.f);

    numberOfStrokePnts = 0;

    // get the stroke points
    vector< UserStroke::strokePoints > points = stroke.getPoints();

    if ( points.size()>0 )
    {
        // the selected triangles are all on the same connected components.
        unsigned int v = vertexCloseToPoint(points[0].p, 0.3);
        strokeConnectCompIndice = vertexConnectedComp[v];
    }

    // faces are marked as follow:
    // -1 : not viewed
    // 0 : background
    // 1 : unclassied
    // 2 : foreground
    // 3 : currently being classified as B, F, U
    vector< int > viewedFaces;
    for (unsigned int i = 0; i < triangles.size(); i++)
        viewedFaces.push_back(-1);

    // the second nb of the pair is the indice of closest point..
    vector< unsigned int > unsortedFaces;

    initializeElementsToClassify(numberOfStrokePnts, points, viewedFaces, unsortedFaces);

    // classify all the unsorted faces
    classifyUnsortedFaces(viewedFaces, unsortedFaces);
    for (unsigned int i = 0; i < viewedFaces.size(); i++)
    {
        faceInSelectionK[i] = -1.f;
        if (SEE_SEEDS)
        {
            if (trianglesUnderMouse[i] == 1.f)
                faceInSelectionK[i] = -4.f;
            else if (viewedFaces[i] == 2)
                faceInSelectionK[i] = 1.f;
        } else {
            if (viewedFaces[i] == 2)
                faceInSelectionK[i] = 1.f;
        }
    }
    numberOfStrokePnts = points.size();

    if (SPY_TIME)
        timing_local.push_back( make_pair(nb_of_local_sel, timer.elapsed()) );
}

void PatchSelection::updateSelection(UserStroke & stroke){
    Trace(3, "PatchSelection::updateSelection ");
    QTime timer;
    if (SPY_TIME)
        timer.start();

    vector< UserStroke::strokePoints > strokeP = stroke.getPoints();

    vector< int > viewedFaces;
    for (unsigned int i = 0; i < faceInSelectionK.size(); i++)
    {
        if (faceInSelectionK[i] == 1.f)
            viewedFaces.push_back(2);
        else
            viewedFaces.push_back(-1);
    }

    vector<unsigned int> unsortedFaces;
    initializeElementsToClassify(numberOfStrokePnts, strokeP, viewedFaces, unsortedFaces);
    classifyUnsortedFaces(viewedFaces, unsortedFaces);

    for (unsigned int i = 0; i < viewedFaces.size(); i++)
    {
        faceInSelectionK[i] = -1.f;
        if (SEE_SEEDS)
        {
            if (trianglesUnderMouse[i] == 1.f)
                faceInSelectionK[i] = -4.f;
            else if (viewedFaces[i] == 2)
                faceInSelectionK[i] = 1.f;
        } else {
            if (viewedFaces[i] == 2)
                faceInSelectionK[i] = 1.f;
        }
    }
    numberOfStrokePnts = strokeP.size();
    if (SPY_TIME)
        timing_local.push_back( make_pair( nb_of_local_sel, timer.elapsed()) );
}

void PatchSelection::refineSelection(UserStroke& _stroke){
    Trace(3, "PatchSelection::refineSelection ");
    vector< UserStroke::strokePoints > points = _stroke.getPoints();

    vector< int > viewedFaces;
    for (unsigned int i = 0; i < triangles.size(); i++)
        viewedFaces.push_back(-1);

    for (unsigned int i = numberOfRefineStrokePnts ; i < points.size(); i++)
    {
        findTrianglesInsideSphere(points[i].p, points[i].r, viewedFaces, 2, 6);
    }
    float colorValue = -1;
    for (unsigned int i = 0; i < triangles.size(); i++){
        if (viewedFaces[i] == 6 && faceInSelectionK[i] != -1)
            colorValue = faceInSelectionK[i];
    }
    if (colorValue != -1){
    for (unsigned int i = 0; i < triangles.size(); i++)
        if (viewedFaces[i] == 6 )
            faceInSelectionK[i] = colorValue;
    }
    numberOfRefineStrokePnts = points.size();
}

void PatchSelection::endSelection(UserStroke & stroke)
{
    Trace(3, "PatchSelection::endSelection");
    QTime timer;
    if (SPY_TIME)
        timer.start();

    updateSelection(stroke);

    verticeInSelection.clear();
    for (unsigned int i = 0; i < vertices.size(); i++)
        verticeInSelection.push_back(-1);

    for (unsigned int i = 0; i < faceInSelectionK.size(); i++)
    {
        if (faceInSelectionK[i] == 1 || faceInSelectionK[i] == -4)
            for (unsigned int j = 0; j < 3; j++)
                verticeInSelection[triangles[i].getVertex(j)]=1;
    }
    setReference();
    if ( refVertex < vertices.size() )
        computeSC(refVertex);

    if (SPY_TIME){
        timing_local.push_back(make_pair( nb_of_local_sel, timer.elapsed()));
        nb_of_local_sel++;
    }
}

void PatchSelection::eraseSelection(UserStroke & stroke){
    Trace(3, "PatchSelection::eraseSelection");

    vector<UserStroke::strokePoints> bgPoints = stroke.getBGPoints();
    vector< int > viewedFaces;
    for (unsigned int i = 0; i < triangles.size(); i++)
        viewedFaces.push_back(-1);

    for (unsigned int i = numberOfBackGroundPnts ; i < bgPoints.size(); i++)
    {
        findTrianglesInsideSphere(bgPoints[i].p, bgPoints[i].r, viewedFaces, 2, 6);
    }
    for (unsigned int i = 0; i < triangles.size(); i++)
        if (viewedFaces[i] == 6 )
            faceInSelectionK[i] = -1.f;
    numberOfBackGroundPnts = bgPoints.size();
}


void PatchSelection::clearSelections(){
    Trace(3, "PatchSelection::clearSelections");
    for (unsigned int i = 0; i<triangles.size(); i++)
        faceInSelectionK[i] = -1.f;
    for (unsigned int i = 0; i < similarity_dist.size(); i++)
        similarity_dist[i] = -1.f;
    refSC.clearValues();
    for (unsigned int i = 0; i < _shapeContexts.size(); i++)
        _shapeContexts[i].clearValues();
    numberOfStrokePnts = 0;
    numberOfBackGroundPnts = 0;
    numberOfRefineStrokePnts = 0;
    trianglesUnderMouse.clear();
    ICPpointsRef.clear();
    ICPpointsRefTransformed.clear();
    for (unsigned int i = 0; i < 3; i++)
    {
        PCAeigenVec[i] = Vec3Df(0.f, 0.f, 0.f);
        PCAeigenVecDeformed[i] = Vec3Df(0.f, 0.f, 0.f);
        PCAeigenVecRef[i] = Vec3Df(0.f, 0.f, 0.f);
    }
}


//***************************************************
//                  EXPANSION
//***************************************************

void PatchSelection::findSimilarSelections(UserStroke& stroke){
    Trace(2, "PatchSelection::findSimilarSelections ");
    QTime timer;
    if (SPY_TIME)
        timer.start();

    if (_similarityAlgo == SVM || _similarityAlgo == ICP)
        _svm.clear();

    strokeForSimiliarity.clear();
    strokeForSimiliarity = stroke.getPoints();

    QTime timer_sc;
    if (SPY_TIME)
        timer_sc.start();

    computeShapeContexts();
    Trace(1, " SC computation time " << timer_sc.elapsed() << "ms ");
    if (SPY_TIME)
    {
        timing_sim t;
        t.time = timer_sc.elapsed();
        t.sel_nb = nb_of_global_sel;
        t.sim_nb = 0;
        timing_SC.push_back(make_pair(diskRadius, t) );
    }

    const QString fileName = "svmtrainerdata";
    std::vector<unsigned int > vertices_ind_ref;
    std::vector<Vec3Df> frame;
    Vec3Df center;
    QTime timer_svm_train;

    switch( _similarityAlgo)
    {
    case BASIC_ALGO:
        break;
    case SVM:
        trainSVM();
        break;
    case EXPORT_SVM:
        writeSVM(fileName, refVertex);
        break;
    case ICP:
        initICPref(vertices_ind_ref, frame, center);
        break;
    case ICP_SVM:
        initICPref(vertices_ind_ref, frame, center);
        if (SPY_TIME)
            timer_svm_train.start();
        initICP_SVM(vertices_ind_ref, frame, center);
        if (SPY_TIME){
            timing_sim t;
            t.time =  timer_svm_train.elapsed();
            t.sel_nb = nb_of_global_sel;
            t.sim_nb = 0;
            timing_SVM_train.push_back(t);
            Trace (1, " timer SVM ICP t " << t.time);
        }
        break;
    default:
        Trace(1, "issue with _similarity algo");
        break;
    }

    // Find the most similar centers
    vector< pair<float, unsigned int> >sortedSim;
    for (unsigned int i = 0; i < similarity_dist.size(); i++)
    {
        sortedSim.push_back(make_pair(similarity_dist[i], i));
    }
    sort(sortedSim.begin(), sortedSim.end() ); // ordre croissant.
    float min = sortedSim[0].first;
    unsigned int decil_max_ind = sortedSim.size()*9/10;
    float max = sortedSim[decil_max_ind].first;
    for (unsigned int i = 0; i < similarity_dist.size(); i++)
    {
        if (similarity_dist[i] >= max )
            similarity_dist[i] = 0.99f;
        else
            similarity_dist[i] = (similarity_dist[i]-min)/(max-min);
    }

    bool stop = false;
    float selection_sim;
    Trace(1, " before while "<< simSelectionsNb);
    simSelectionsNb = 0;
    while(!stop && simSelectionsNb < nb_of_similar_sel)
    {
        Trace(1, " in while "<< simSelectionsNb);
        selection_sim = selectKiemSimilarPatch(sortedSim, vertices_ind_ref);
        if (selection_sim > SIMILARITY_THRESHOLD)
            stop = true;
        else
            simSelectionsNb++;
    }
    if (SPY_TIME)
        timing_global.push_back(make_pair(nb_of_global_sel,timer.elapsed()));
    nb_of_global_sel++;
}

void PatchSelection::changeSimiliarityThreshold(int val){
    nb_of_similar_sel =  val;
    Trace(2, " PatchSelection::changeSimilarityThreshold to " << nb_of_similar_sel);
    // reset all the faces that are not part of the ref patch to 0
    for (unsigned int i = 0; i < faceInSelectionK.size(); i++)
        if (faceInSelectionK[i] != 1.f && faceInSelectionK[i] != -4.f)
            faceInSelectionK[i] = -1.f;

    vector< pair<float, unsigned int> >sortedSim;
    for (unsigned int i = 0; i < similarity_dist.size(); i++)
    {
        sortedSim.push_back(make_pair(similarity_dist[i], i));
    }
    // sort the distances of all the vertice to the ref
    sort(sortedSim.begin(), sortedSim.end() );

    bool stop = false;
    float selection_sim;
    simSelectionsNb = 0;

    std::vector<unsigned int> vertices_ind_ref;
    // find all the vertices inside the radius
    findVerticesInsideDisk(vertices_ind_ref, refVertex);
    while(!stop && simSelectionsNb < nb_of_similar_sel)
    {
        // find the k° similar patch
        selection_sim = selectKiemSimilarPatch(sortedSim, vertices_ind_ref);
        if (selection_sim > SIMILARITY_THRESHOLD)
            stop = true;
        else
            simSelectionsNb++;
    }
}

void PatchSelection::setSimilarityAlgo(PATCH_SIM_ALGO algo, UserStroke &stroke){
    for (unsigned int i = 0; i < faceInSelectionK.size(); i++)
        if (faceInSelectionK[i] != 1.f && faceInSelectionK[i] != -4.f)
            faceInSelectionK[i] = -1.f;
    _similarityAlgo = algo;
    findSimilarSelections(stroke);
}


void PatchSelection::openSVMselection(const QString & filename){

    Trace(1, "PatchSelection::openSVMselection");
    QString number = filename;
    number.chop(8);

    QString entete ="/home/guy/builds/libsvm-3.17/tools/SVMtestDATA";
    unsigned int nbDigit = number.size()-entete.size();
    QString nb = number.right(nbDigit);
    unsigned int vertexNumber = nb.toUInt();
    Trace(1, "vertex nb" << vertexNumber );


    const std::string file_n = qPrintable(filename);
    ifstream in (file_n.c_str ());
    if (!in)
       exit (EXIT_FAILURE);

    Trace(1, "in open" );
    QString indiceFileS = "./SVMtestDATAIND";
    indiceFileS.append(nb);
    const std::string file_ind_n = qPrintable(indiceFileS);
    Trace(1, " indice File Name "<< file_ind_n);
    ifstream inIndice (file_ind_n.c_str ());
    if (!inIndice)
       exit (EXIT_FAILURE);

    unsigned int cur_v;
    int selected;
    std::vector<int> selectedV;
    for (unsigned int i = 0; i < vertices.size(); i++)
        selectedV.push_back(-1.f);
    while (inIndice.good())
    {
        inIndice >> cur_v;
        in >> selected;
        selectedV[cur_v] = selected;

        Trace(1, "cur_v "<< cur_v<< " in " << selected);
    }
    in.close();
    inIndice.close();
    bool triangleSelected;
    for (unsigned int i = 0; i< faceInSelectionK.size(); i++)
    {
        triangleSelected = true;
        for (unsigned int j = 0; j < 3; j++)
        {
            cur_v = triangles[i].getVertex(j);
            if (selectedV[cur_v] == -1)
                triangleSelected = false;
        }
        if (triangleSelected)
        {
            Trace(1, " triangle " << i << "selected ");
            faceInSelectionK[i] = 1.5f;
        }
    }
 }

void PatchSelection::saveSelectionHistImage(){
    Trace(3, "PatchSelection::saveSelectionHistImage");
    QPixmap image;

    vector<bool> verticeSelected;
    unsigned int nb_selected_v = 0;
    for (unsigned int i = 0; i < vertices.size();i++)
        verticeSelected.push_back(false);
    unsigned int cur_v;
    for (unsigned int i = 0; i < faceInSelectionK.size(); i++)
    {
        if (faceInSelectionK[i] == 1)
        {
            for (unsigned int j = 0; j < 3; j++)
            {
                cur_v = triangles[i].getVertex(j);
                if (!verticeSelected[cur_v])
                    nb_selected_v++;
                verticeSelected[cur_v] = true;
            }
        }
    }
    unsigned int nb_of_values = _shapeContexts[refVertex].getBinsNumber();
    const int cell_height =  15;
    image = QPixmap(cell_height*nb_of_values, cell_height*nb_selected_v);
    int i_pos=0;
    int j_pos=0;

    QPainter painter(&image);
    painter.setPen(Qt::black);
    painter.setBrush(Qt::white);

    std::vector<double> current_hist_values;
    for (unsigned int i = 0; i < vertices.size();i++)
    {
        if (verticeSelected[i])
        {
            current_hist_values.clear();
            _shapeContexts[i].getValues(current_hist_values);
            double max_value = 0.f;
            double min_value = FLT_MAX;
            for (unsigned int j = 0; j < current_hist_values.size(); j++)
            {
                max_value = max(max_value, current_hist_values[j]);
                min_value = min(min_value, current_hist_values[j]);
            }
            float max_min = max_value-min_value;
            for (unsigned int j = 0; j < current_hist_values.size(); j++)
            {
                double val;
                if (max_min != 0)
                    val = (current_hist_values[j]-min_value)/max_min;
                else
                    val = current_hist_values[j];
                val*=255;
                QColor color(val*0.8, 255-val*0.8, 255-0.8*val);
                color.toHsv();
                color.setHsv(color.hue(), color.saturation()*3/4, color.value());
                painter.fillRect(j_pos, i_pos, cell_height, cell_height, color);
                j_pos += cell_height;
            }
            i_pos += cell_height;
            j_pos = 0;
        }
    }
    QFile file("yourFile.png");
    file.open(QIODevice::WriteOnly);
    image.save(&file, "PNG");

}

void PatchSelection::drawICPDebug(){

    if (_similarityAlgo == ICP_SVM)
    {
        glColor4f(0.0f, 1.f, 0.2f, 0.5f);
        if (refVertex < vertices.size())
        {
            // center of the frame
            Vec3Df p = vertices[refVertex].getPos();
            BasicGL::drawSphere(p[0], p[1], p[2], 0.01, 30, 30);
            Vec3Df center = p;

            // vectors of the ref frame
            glColor3f(1.0, 0.f, 0.f);
            glBegin(GL_LINES);
            glVertexVec3Df(center);
            glVertexVec3Df(center+PCAeigenVecRef[0]/5.);
            glEnd();
            glColor3f(0.0, 1.f, 0.f);
            glBegin(GL_LINES);
            glVertexVec3Df(center);
            glVertexVec3Df(center+PCAeigenVecRef[1]/5.);
            glEnd();
            glColor3f(0.0, 0.f, 1.f);
            glBegin(GL_LINES);
            glVertexVec3Df(center);
            glVertexVec3Df(center+PCAeigenVecRef[2]/5.);
            glEnd();

            // center and vectors of the local frame
            if (ind_vertex < vertices.size())
            {
                center = vertices[ind_vertex].getPos();
                glColor3f(1.0, 0.f, 0.f);
                glBegin(GL_LINES);
                glVertexVec3Df(center);
                glVertexVec3Df(center+PCAeigenVec[0]/5.);
                glEnd();
                glColor3f(0.0, 1.f, 0.f);
                glBegin(GL_LINES);
                glVertexVec3Df(center);
                glVertexVec3Df(center+PCAeigenVec[1]/5.);
                glEnd();
                glColor3f(0.0, 0.f, 1.f);
                glBegin(GL_LINES);
                glVertexVec3Df(center);
                glVertexVec3Df(center+PCAeigenVec[2]/5.);
                glEnd();
            }

            // to draw the points of the reference frame after and before ICP
            // useful for debug but bad for the frame rate!

            //        glColor4f(1.0f, 0.f, 0.2f, 0.5f);
            //        unsigned int i = 0;
            //        while (i < ICPpointsRefTransformed.size())
            //        {
            //            Vec3Df p = ICPpointsRefTransformed[i];
            //            BasicGL::drawSphere(p[0], p[1], p[2], 0.002, 30, 30);
            //            i = i+4;
            //        }

            //        glColor4f(0.0f, 0.f, 0.9f, 0.5f);
            //        i = 0;
            //        while (i < ICPpointsRef.size())
            //        {
            //            Vec3Df p = ICPpointsRef[i];
            //            BasicGL::drawSphere(p[0], p[1], p[2], 0.003, 30, 30);
            //            i = i+4;
            //        }

        }
    }
    else if (_similarityAlgo == ICP){
        glColor4f(1.0f, 0.6f, 0.0f, 1.f);
        unsigned int i = 0;
        while (i < ICPpointsRefTransformed.size())
        {
            Vec3Df p = ICPpointsRefTransformed[i];
            BasicGL::drawSphere(p[0], p[1], p[2], 0.003, 30, 30);
            i = i+1;
        }

        glColor4f(0.0f, 0.69f, 0.09f, 1.0f);
        i = 0;
        while (i < ICPpointsRef.size())
        {
            Vec3Df p = ICPpointsRef[i];
            BasicGL::drawSphere(p[0], p[1], p[2], 0.003, 30, 30);
            i = i+1;
        }
        if (refVertex < vertices.size())
        {
            Vec3Df p = vertices[refVertex].getPos();
            BasicGL::drawSphere(p[0], p[1], p[2], 0.01, 30, 30);
            Vec3Df center = p;
            glColor3f(1.0, 0.f, 0.f);
            glBegin(GL_LINES);
            glVertexVec3Df(center);
            glVertexVec3Df(center+PCAeigenVecRef[0]);
            glEnd();
            glColor3f(0.0, 1.f, 0.f);
            glBegin(GL_LINES);
            glVertexVec3Df(center);
            glVertexVec3Df(center+PCAeigenVecRef[1]);
            glEnd();
            glColor3f(0.0, 0.f, 1.f);
            glBegin(GL_LINES);
            glVertexVec3Df(center);
            glVertexVec3Df(center+PCAeigenVecRef[2]);
            glEnd();

            if (ind_vertex < vertices.size())
            {

                glDisable(GL_DEPTH_TEST);
                center = vertices[ind_vertex].getPos();
                glColor3f(1.0, 0.f, 0.f);
                glBegin(GL_LINES);
                glVertexVec3Df(center);
                glVertexVec3Df(center+PCAeigenVec[0]);
                glEnd();
                glColor3f(0.0, 1.f, 0.f);
                glBegin(GL_LINES);
                glVertexVec3Df(center);
                glVertexVec3Df(center+PCAeigenVec[1]);
                glEnd();
                glColor3f(0.0, 0.f, 1.f);
                glBegin(GL_LINES);
                glVertexVec3Df(center);
                glVertexVec3Df(center+PCAeigenVec[2]);
                glEnd();

                glEnable(GL_DEPTH_TEST);
                glColor4f(0.0f, 0.79f, 0.19f, 1.0f);
                BasicGL::drawSphere(center[0], center[1], center[2], 0.01, 30, 30);

                glColor4f(1.0f, 0.7f, 0.0f, 1.f);
                BasicGL::drawSphere(centerVecDeformed[0], centerVecDeformed[1], centerVecDeformed[2], 0.01, 30, 30);

                glDisable(GL_DEPTH_TEST);
                glLineWidth(3.f);
                glColor3f(1.0, 0.f, 0.f);
                glBegin(GL_LINES);
                glVertexVec3Df(centerVecDeformed);
                glVertexVec3Df(centerVecDeformed+0.5*PCAeigenVecDeformed[0]);
                glEnd();
                glColor3f(0.0, 1.f, 0.f);
                glBegin(GL_LINES);
                glVertexVec3Df(centerVecDeformed);
                glVertexVec3Df(centerVecDeformed+0.5*PCAeigenVecDeformed[1]);
                glEnd();
                glColor3f(0.0, 0.f, 1.f);
                glBegin(GL_LINES);
                glVertexVec3Df(centerVecDeformed);
                glVertexVec3Df(centerVecDeformed+0.5*PCAeigenVecDeformed[2]);
                glEnd();
                glLineWidth(1.f);
                glEnable(GL_DEPTH_TEST);

            }
        }
    }
}


void PatchSelection::drawShapeContextDebug(){
    glColor4f(0.0f, 1.f, 0.2f, 0.5f);
    if (refVertex < vertices.size())
    {
        Vec3Df p = vertices[refVertex].getPos();
        BasicGL::drawSphere(p[0], p[1], p[2], 0.01, 30, 30);
    }
}


//***************************************************
//               SELECTION (PROTECTED)
//***************************************************
void PatchSelection::initializeElementsToClassify( unsigned int starting_point,
                                                   vector< UserStroke::strokePoints>& points,
                                                  std::vector< int > & viewedFaces,
                                                  std::vector< unsigned int > & unsortedFaces){
    Trace(3, "PatchSelection::initializeElementsToClassify");

    // We find all the triangles under the stroke
    // ie : inside the spheres of center : points.p and radius points.r
    bool strokeOutCurSelection = false;
    for (unsigned int i = starting_point ; i < points.size(); i++)
    {
        // 2 : already selected faces.
        // 6 : faces not previously selected which are inside the sphere radius
        bool curPntOut = findTrianglesInsideSphere(points[i].p, points[i].r, viewedFaces, 2, 6);
        if (curPntOut)
            strokeOutCurSelection = true;
    }
    // we do something only if one points of the new part of the stroke
    // is outside the current selection
    if (strokeOutCurSelection)
    {
        // something smarter to do..
        // do we use the patchCentroid, normal, and PCApoints of all the strokes or just of the last one ?
        vector<Vec3Df> pointsForPCA;
        Vec3Df strokeCentroid = Vec3Df(0.f, 0.f, 0.f);
        strokeCentroid = Vec3Df(0.f, 0.f, 0.f);
        Vec3Df c_n = Vec3Df(0.f, 0.f, 0.f);

        // for all the NEW selected faces.
        vector<bool> insideVertices;
        for (unsigned int i = 0; i < vertices.size(); i++)
            insideVertices.push_back(false);
        for (unsigned int i = 0; i < viewedFaces.size(); i++)
        {
            if (viewedFaces[i] == 6)
            {
                for (unsigned int j = 0; j < 3; j++)
                {
                    unsigned int v = triangles[i].getVertex(j);
                    if (!insideVertices[v])
                    {
                        pointsForPCA.push_back(vertices[v].getPos());
                        strokeCentroid += vertices[v].getPos();
                    }
                }
                c_n+= triangles[i].getSmoothNorm();
                viewedFaces[i] = 2;
            }
        }
        strokeCentroid/= (float)(pointsForPCA.size());
        c_n.normalize();

        float eigenVal[3];
        Vec3Df eigenVec[3];
        // we compute the PCA of current selected faces
        principalComponentAnalysis(pointsForPCA, eigenVal, eigenVec);

        // the eigenVector of the PCA are used to determined a bounding box, to defined where to stop;
        Vec3Df t_xyz, t_pca;
        float radius_max_v = 0.f;
        float radius_max_u = 0.f;
        float radius_max_w = 0.f;
        for (unsigned int i = 0; i < pointsForPCA.size(); i++)
        {
            t_xyz = pointsForPCA[i]-strokeCentroid;
            t_pca = t_xyz.expressInBasis(eigenVec[0], eigenVec[1], eigenVec[2]);
            radius_max_v = max ( radius_max_v, t_pca[2]);
            radius_max_u = max ( radius_max_u, t_pca[1]);
            radius_max_w = max ( radius_max_w, t_pca[0]);
        }

        float radius_med = 0;
        for (unsigned int i = starting_point; i < points.size(); i++)
            radius_med += points[i].r;
        radius_med /= (points.size()-starting_point);

        radius_max_v+= 2*radius_med;
        radius_max_u+= 4*radius_med;

        // we perform a flood filling algorithm to find
        // all the faces that are unclassified..
        queue<unsigned int> toClassify;
        for (unsigned int i = 0; i < viewedFaces.size(); i++)
            if (viewedFaces[i] == 3)
                toClassify.push(i);

        unsigned int cur_face;
        while (!toClassify.empty())
        {
            cur_face = toClassify.front();
            toClassify.pop();
            t_xyz = getBar(cur_face)-strokeCentroid;

            Vec3Df n_i= triangles[cur_face].getSmoothNorm();
            float cos_angle = Vec3Df::dotProduct(n_i, c_n);

            t_pca = t_xyz.expressInBasis(eigenVec[0], eigenVec[1], eigenVec[2]);
            if ( ( (abs(t_pca[2]) > radius_max_v) ||
                 (abs(t_pca[1]) > radius_max_u) ||
                 (cos_angle < 0) ) && (viewedFaces[cur_face] !=2)  )
                viewedFaces[cur_face] = 0;
            else {
                unsortedFaces.push_back(cur_face);
                viewedFaces[cur_face] = 1;
                for (unsigned int i = 0; i < faceNeighbors[cur_face].size(); i++)
                {
                    unsigned int next_f =  faceNeighbors[cur_face][i];
                    if (viewedFaces[next_f] == -1)
                    {
                        toClassify.push(next_f);
                        viewedFaces[next_f]= 3;
                    }
                }
            }
        }
    }
}

bool PatchSelection::findTrianglesInsideSphere( Vec3Df center, float radius,
                                                std::vector<int> &  inside,
                                                int inside_value,
                                                int new_inside_value)
{
    Trace(3, "PatchSelection::findTrianglesInsideSphere");
    bool found = false;
    unsigned int i, nei_f, cur_f;
    queue< unsigned int > toClassify;
    float dist ;

    i = 0;
    // we search a triangle inside the sphere that is not already inside the selection
    while ( !(found || i >=triangles.size() ))
    {
        dist = (getBar(i)-center).getLength();
        if ( (dist<=radius) &&
             (vertexConnectedComp[triangles[i].getVertex(0)] == strokeConnectCompIndice) &&
             inside[i] != inside_value)
            found = true;
        else
            i++;
    }

    // if we found one we set it to the new inside value
    // and fill the "classify" queue with it neighbors
    if (found)
    {
        inside[i] = new_inside_value;
        for (unsigned int j = 0; j < faceNeighbors[i].size(); j++)
        {
            nei_f = faceNeighbors[i][j];
            if (inside[nei_f] != new_inside_value
                    && inside[nei_f] != inside_value
                    && inside[nei_f] != 4)
            {
                toClassify.push(nei_f);
                inside[nei_f] = 4;
            }
        }
        while (!toClassify.empty())
        {
            cur_f = toClassify.front();
            toClassify.pop();
            dist = (getBar(cur_f)-center).getLength();
            if (dist < radius
                    && (vertexConnectedComp[triangles[i].getVertex(0)] == strokeConnectCompIndice))
            {
                inside[cur_f] = new_inside_value;
                for (unsigned int f = 0; f < faceNeighbors[cur_f].size(); f++)
                {
                    nei_f = faceNeighbors[cur_f][f];
                    if (inside[nei_f] != new_inside_value && inside[nei_f] != 4)
                    {
                        toClassify.push(nei_f);
                        inside[nei_f] = 4;
                    }
                }
            } else
                 inside[cur_f] = 3;
        }
    }

    for (unsigned int i = 0; i < inside.size(); i++)
        if (inside[i] == new_inside_value && new_inside_value != 5)
            trianglesUnderMouse[i] = 1.f;
    return found;
}

void PatchSelection::classifyUnsortedFaces(std::vector< int > & faces,
                                           std::vector< unsigned int> const & unsortedFaces )
{
    Trace(3, "PatchSelection::classifyUnsortedFaces");
    unsigned int cur_f, nei_f;
    Vec3Df bar_f, norm_f, bar_nei, norm_nei;
    float d;
    // to keep a track of the triangles that are currently being sort
    // (ie : that are in the facesToSort vector)
    vector<bool> facesBeingSorted;
    for (unsigned int i = 0; i < triangles.size(); i++)
        facesBeingSorted.push_back(false);

    // faces that need to be sort, and that are neighbors of either a BG face, or a FG face
    std::multimap< float, pair<int, unsigned int> > facesToSort;

    // init of facesToSort

    float min_dist;
    unsigned int FG_BG;
    for (unsigned int i = 0; i < unsortedFaces.size(); i++)
    {
        cur_f = unsortedFaces[i];
        bar_f = getBar(cur_f);
        norm_f = triangles[cur_f].getSmoothNorm();
        min_dist = FLT_MAX;
        for (unsigned int j = 0; j < faceNeighbors[cur_f].size(); j++)
        {
            nei_f = faceNeighbors[cur_f][j];
            // if a neighbor of the face is part of the BG or FG,
            // we compute the cost of adding the current face in one of the competitors set
            // and either we add the current face to the facesToSort list (if it isn't already inside)
            // either we update it cost, and competitor set (if necessary)
            if (faces[nei_f] == 0 || faces[nei_f] == 2)
            {
                bar_nei = getBar(nei_f);
                norm_nei = triangles[nei_f].getSmoothNorm();
                if (LcrossDistUsed)
                    d = distanceL2_cross( bar_f,bar_nei , norm_f, norm_nei, angle);
                else
                    d = distanceL2_1w(bar_f, bar_nei, norm_f, norm_nei, angle);

                if (d < min_dist)
                {
                    min_dist = d;
                    FG_BG = faces[nei_f];
                }
            }
        }
        // if the face cur_f is a neighbor of the FG or BG
        if (min_dist != FLT_MAX)
        {
            facesToSort.insert(make_pair(min_dist, make_pair(FG_BG, cur_f)));
            facesBeingSorted[cur_f] = true;
        }
    }
    std::multimap<float, pair<int, unsigned int> >::iterator it;

    // while their is faces to sort
    while (! facesToSort.empty() )
    {
        it = facesToSort.begin();
        // new face sorted
        cur_f = (*it).second.second;
        faces[cur_f] = (*it).second.first;
        facesToSort.erase(it);
        facesBeingSorted[cur_f] = false;
        bar_f = getBar(cur_f);
        norm_f = triangles[cur_f].getSmoothNorm();

        for (unsigned int i = 0; i < faceNeighbors[cur_f].size(); i++)
        {
            nei_f = faceNeighbors[cur_f][i];
            if (faces[nei_f] == 1)
            {
                bar_nei = getBar(nei_f);
                norm_nei = triangles[nei_f].getSmoothNorm();

                if (LcrossDistUsed)
                    d = distanceL2_cross( bar_f, bar_nei, norm_f, norm_nei, angle);
                else
                    d = distanceL2_1w( bar_f, bar_nei, norm_f, norm_nei, angle);

                // if the neighbor is already in the facesToSort list we only update it
                if (facesBeingSorted[nei_f])
                {
                    it = facesToSort.begin();
                    bool found = false;
                    while ( !(found || it == facesToSort.end()))
                    {
                        if ((*it).second.second == nei_f)
                            found = true;
                        else
                            it++;
                    }
                    assert(found);
                    if (d < (*it).first)
                    {
                        facesToSort.erase(it);
                        facesToSort.insert(make_pair(d, make_pair( faces[cur_f], nei_f)));
                    }
                } else {
                    // else we add it
                    facesToSort.insert(make_pair(d, make_pair(faces[cur_f], nei_f)));
                    facesBeingSorted[nei_f] = true;
                }
            }
        }
    }
}

void PatchSelection::findVerticesInsideDisk( std::vector<unsigned int> &vertices_ind, unsigned int v_indice )
{
    vertices_ind.clear();
    std::vector<bool> viewed;
    for (unsigned int i = 0; i < vertices.size(); i++)
        viewed.push_back(false);

    std::queue<unsigned int> toLook;

    viewed[v_indice] = true;
    vertices_ind.push_back(v_indice);

    for (unsigned int i = 0; i < oneRing[v_indice].size(); i++ )
    {
        toLook.push(oneRing[v_indice][i]);
        viewed[oneRing[v_indice][i]] = true;
    }

    unsigned int cur_v;
    float dist;
    Vec3Df centerPos = vertices[v_indice].getPos();
    while (!toLook.empty())
    {
        Trace(3, " in to Look while ");
        cur_v = toLook.front();
        toLook.pop();
        dist = (vertices[cur_v].getPos()-centerPos).getLength();
        if (dist < diskRadius)
        {

            for (unsigned int i = 0; i < oneRing[cur_v].size(); i++)
            {
                unsigned int nei = oneRing[cur_v][i];
                if (!viewed[nei])
                {
                    viewed[nei] = true;
                    toLook.push(nei);
                }
            }
            vertices_ind.push_back(cur_v);
        }
    }
}


//***************************************************
//                  EXPANSION
//***************************************************

void PatchSelection::recomputeSimilarityDist(){
    Trace(2, "PatchSelection::recomputeSimilarityDist");
    float max = FLT_MIN;
    float min = FLT_MAX;

    for (unsigned int i = 0; i < similarity_dist.size(); i++)
    {
        switch (histoDist){
        case L2:
            similarity_dist[i] = refSC.distanceL2(_shapeContexts[i]);
            break;
        case L1:
             similarity_dist[i] = refSC.distanceL1(_shapeContexts[i]);
            break;
        case CHI2:
            similarity_dist[i] = refSC.distanceChi2(_shapeContexts[i]);
            break;
        case CHI2N:
            similarity_dist[i] = refSC.distanceChi2Normalized(_shapeContexts[i]);
            break;
        default:
            break;
        }
        if (similarity_dist[i] > max)
            max = similarity_dist[i];
        if (similarity_dist[i] < min)
            min = similarity_dist[i];
    }

    float max_min = max-min;
    for (unsigned int i = 0; i< similarity_dist.size(); i++)
        similarity_dist[i] = (similarity_dist[i]-min)/max_min;
}


void PatchSelection::setReference(){
    Trace(1, "PatchSelection::setReference");
    miniball(verticeInSelection, 1, vertices, refCenter, diskRadius, 0.05);

    int v = vertexCloseToPoint(refCenter, 0.1);
    if (v < 0 || v >= (int) vertices.size())
    {
        Trace(1, " error" );
        v = vertexCloseToPoint(refCenter, 0.9);
    }
    refVertex = v;
    diskRadius*=1.2f;
    Trace(1, " diskRadius" << diskRadius);
}

float PatchSelection::selectKiemSimilarPatch( const std::vector<std::pair<float, unsigned int> > & sortedSim,
                                              const std::vector<unsigned int> & vertices_ind_ref){

    unsigned int nnv_ind = 0;
    unsigned int nnv = sortedSim[nnv_ind].second;
    bool stop = false;
    bool closeToCurSel;

    // find the first vertex which is similar to the selection center, and not already inside an existing selection
    while (! (stop || nnv_ind >= vertices.size() ) )
    {
        closeToCurSel = false;
        for (unsigned int i = 0; i < oneRingTr[nnv].size(); i++)
        {
            if (faceInSelectionK[oneRingTr[nnv][i]] != -1.f)
                closeToCurSel = true;
            if ( (vertices[nnv].getPos()-refCenter).getLength()<diskRadius*2.f/3.f)
                closeToCurSel = true;
        }
        // if the current vertex is too close to the refence patch
        // we keep looking
        if (closeToCurSel)
        {
            nnv_ind++;
            nnv = sortedSim[nnv_ind].second;
        }
        else
            stop = true;
    }

    if ( nnv < vertices.size())
    {
        const QString fileName = "SVMtestDATA";
        std::vector<unsigned int> seed;
        std::vector<unsigned int> negSeed;
        seed.push_back(nnv);
        std::vector<unsigned int> vertices_ind;
        Matrix firstTransform(4,4);
        Matrix icpTransform(4,4);
        std::vector<Vec3Df> loc_frame_icp;
        Vec3Df centerICP;
        switch( _similarityAlgo)
        {
        case BASIC_ALGO:
            basicSimilarityAlgo(seed, negSeed, nnv);
            break;
        case SVM:
            predictSVM(nnv);
            break;
        case EXPORT_SVM:
            writeSVM(fileName, nnv);
            break;
        case ICP_SVM:
            computeICP_SVM(nnv, vertices_ind_ref);
            break;
        case ICP:
            computeICP(nnv, vertices_ind_ref, vertices_ind, firstTransform, icpTransform, loc_frame_icp, centerICP);
            computeBasicAlgoFromICP(firstTransform, icpTransform );
            break;
        default:
            break;
        }
    }
    return sortedSim[nnv_ind].first;
}

void PatchSelection::computeICP_SVM(unsigned int v_indice,
                                    const std::vector<unsigned int> & vertices_ind_ref){

    // COMPUTE ICP
    Vec3Df center;
    std::vector<Vec3Df> f;
    std::vector<unsigned int> vertices_ind;
    Matrix firstTransform(4,4);
    Matrix icpTransform(4,4);

    computeICP(v_indice, vertices_ind_ref, vertices_ind, firstTransform, icpTransform, f, center);

    QTime timer;
    if (SPY_TIME)
        timer.start();

    // express vectors in feature space for the svm
    std::vector< int > _labels;
    std::vector<std::vector<double> > _vectors;
    getValuesforSVM_withICP(_labels, _vectors, vertices_ind, f, center );

    // PREDICT SVM
    std::vector<std::vector<double> > _probabilities;
    _svm.predictWithProb(_labels, _vectors, _probabilities);

    if (SPY_TIME)
    {
        timing_sim t;
        t.time = timer.elapsed();
        t.sel_nb = nb_of_global_sel;
        t.sim_nb = simSelectionsNb;
        timing_SVM_predict.push_back(t);
    }

    vector<float>probaV;
    for (unsigned int i = 0; i < vertices.size(); i++)
        probaV.push_back(0.f);

    vector<int> selectedV;
    for (unsigned int i = 0; i< vertices.size(); i++)
        selectedV.push_back(-1);

    std::vector<std::pair<float, unsigned int> > probaAndV;
    std::vector<std::pair<float, unsigned int> > negProbaAndV;
    unsigned int countNotNegProb = 0;
    for (unsigned int i = 0; i < _labels.size(); i++)
    {
        if (_labels[i] == 1)
        {
            selectedV[vertices_ind[i]] = 1;
            probaV[vertices_ind[i]] = _probabilities[i][0];
            if (SEE_SVM) {
                probaAndV.push_back(make_pair( _probabilities[i][0],vertices_ind[i] ));
                if (_probabilities[i][0] != 0.f)
                    countNotNegProb++;
            }
        } else {
            selectedV[vertices_ind[i]]=-4.f;
            if (SEE_SVM)
            {
                negProbaAndV.push_back(make_pair( _probabilities[i][0],vertices_ind[i] ));
                Trace(4, " neg Seed " <<  _probabilities[i][0] << " " << vertices_ind[i]  );
            }
        }
    }

    if (SEE_SVM) // useful for debug mode
    {
        sort(probaAndV.begin(), probaAndV.end());
        sort(negProbaAndV.begin(), negProbaAndV.end());

        vector<unsigned int> seed;
        if (!probaAndV.empty())
        {
            for (unsigned int i = probaAndV.size()-1; i > probaAndV.size()*2.f/10.f; i--)
                seed.push_back(probaAndV[i].second);
        }
        vector<unsigned int> negSeed;
        if (!negProbaAndV.empty())
        {
            NegSeedsToDraw.clear();
            for (unsigned int i = 0; i < negProbaAndV.size(); i++)
            {
                negSeed.push_back(negProbaAndV[i].second);
                NegSeedsToDraw.push_back(negProbaAndV[i].second);
            }
        }
        Trace(1, "seed.size " << seed.size() << " nb of proba " << countNotNegProb );
        Trace(1, "NEGseed.size " << NegSeedsToDraw.size() );

        for (unsigned int i = 0; i < selectedV.size(); i++)
            selectedV[i]=-1;
        for(unsigned int i = 0; i < seed.size(); i++)
            selectedV[seed[i]] = 1.f;

        bool triangleSelected;
        unsigned int cur_v;
        for (unsigned int i = 0; i< faceInSelectionK.size(); i++)
        {
            triangleSelected = true;
            for (unsigned int j = 0; j < 3; j++)
            {
                cur_v = triangles[i].getVertex(j);
                if (selectedV[cur_v] == -1)
                    triangleSelected = false;
            }
            if (triangleSelected)
                faceInSelectionK[i] = 1.5;
            else if (faceInSelectionK[i] != 1.f && faceInSelectionK[i] != -4.f)
                faceInSelectionK[i] = -1.;
        }
        basicSimilarityAlgo(seed, negSeed, v_indice);
    } else {
        // set the triangles found by the svm as selected
        // the color of the triangles depends on the probability of its label
        bool triangleSelected;
        unsigned int cur_v;
        float probaT;
        for (unsigned int i = 0; i< faceInSelectionK.size(); i++)
        {
            triangleSelected = true;
            probaT=0.f;
            for (unsigned int j = 0; j < 3; j++)
            {
                cur_v = triangles[i].getVertex(j);
                if (selectedV[cur_v] == -4)
                    triangleSelected = false;
                else if (selectedV[cur_v] == -1 )
                    triangleSelected = false;
                else
                    probaT+= probaV[cur_v];
            }
            if (triangleSelected && probaT > 0.35f)
            {
                faceInSelectionK[i] = 2- probaT/3.f;
            }
        }
    }
}

void PatchSelection::basicSimilarityAlgo( std::vector<unsigned int> seeds,
                                          std::vector<unsigned int> negSeeds,
                                          unsigned int center){
    // faces are marked as follow:
    // -1 : not viewed
    // 0 : background
    // 1 : unclassied
    // 2 : foreground
    // 3 : currently being classified as B, F, U
    vector< int > viewedFaces;
    for (unsigned int i = 0; i < triangles.size(); i++)
        viewedFaces.push_back(-1.f);

    for (unsigned int j = 0; j < seeds.size(); j++)
    {
        for (unsigned int i = 0; i < oneRingTr[seeds[j]].size(); i++)
            viewedFaces[oneRingTr[seeds[j]][i]] = 2.f;
    }
    queue<unsigned int> toClassify;
    for (unsigned int i = 0; i < viewedFaces.size(); i++)
    {
        if (viewedFaces[i] == 2.f)
        {
            for (unsigned int j = 0; j < faceNeighbors[i].size(); j++)
            {
                if (viewedFaces[faceNeighbors[i][j]] == -1.f)
                {
                    viewedFaces[faceNeighbors[i][j]] = 3.f;
                    toClassify.push(faceNeighbors[i][j]);
                }
            }
        }
    }
    unsigned int cur_f;
    float dist;
    // we keep all the unclassified faces in this vector
    vector< unsigned int > unsortedFaces;

    while (!toClassify.empty())
    {
        cur_f = toClassify.front();
        toClassify.pop();
        dist = (vertices[center].getPos()-getBar(cur_f)).getLength();
        if (dist < diskRadius)
        {
            viewedFaces[cur_f] = 1;
            unsortedFaces.push_back(cur_f);
            for (unsigned int i = 0; i < faceNeighbors[cur_f].size(); i++)
            {
                if (viewedFaces[faceNeighbors[cur_f][i]] == -1.f)
                {
                    viewedFaces[faceNeighbors[cur_f][i]] = 3.f;
                    toClassify.push(faceNeighbors[cur_f][i]);
                }
            }
        } else {
            viewedFaces[cur_f] = 0;
        }
    }
    for (unsigned int i = 0 ; i < negSeeds.size(); i++)
        viewedFaces[negSeeds[i]] = 0;

    classifyUnsortedFaces(viewedFaces, unsortedFaces);

    for (unsigned int i = 0; i < viewedFaces.size(); i++)
    {
        if (viewedFaces[i] == 2 && faceInSelectionK[i] != 1.f)
            faceInSelectionK[i] = 1.5f;
    }
}

//***************************************************
//              PATCH SHAPE CONTEXT
//***************************************************

void PatchSelection::computeShapeContexts()
{
    Trace(1, "PatchSelection::computeShapeContexts:");
    int previous_i = 0;
    int cur_i;
#pragma omp parallel for
    for (unsigned int i = 0; i < vertices.size(); i++)
    {
        cur_i =  (i/(float)vertices.size())*100;
        if ( cur_i > previous_i+5)
        {
            emit similarityMapComputed(cur_i);
            previous_i  = cur_i;
        }
        if (similarity_dist[i] == -1.f)
            computeSC(i);
    }
}

float PatchSelection::computeSCangle(unsigned int v, Vec3Df n_ref ){
    Vec3Df n_cur = vertices[v].getNormal();
    return acos(Vec3Df::dotProduct(n_cur, n_ref));
}


void PatchSelection::computeSC(unsigned int v){
    Trace(3, "PatchSelection::computeSC:"<<v);

    // find all the vertice inside a disk of radius diskRadius
    // (same as the findVerticesInsideDisk function)
    // but compute at the same time the value to add in the PSC for each vertex
    std::queue<unsigned int> toLook;
    std::vector<bool> viewed;
    for (unsigned int i = 0; i < vertices.size(); i++)
        viewed.push_back(false);
    viewed[v] = true;
    for (unsigned int i = 0; i < oneRing[v].size(); i++ )
    {
        toLook.push(oneRing[v][i]);
        viewed[oneRing[v][i]] = true;
    }
    unsigned int cur_v;
    float dist;
    Vec3Df centerPos = vertices[v].getPos();
    Vec3Df n_ref = vertices[v].getNormal();

    while (!toLook.empty())
    {
        cur_v = toLook.front();
        toLook.pop();

        dist = (vertices[cur_v].getPos()-centerPos).getLength();
        if (dist < diskRadius)
        {
            // we compute the "area of the vertex
            // so that the PSC is invariant to different mesh triangulation
            float area = 0.f;
            for (unsigned int i = 0; i < oneRingTr[cur_v].size(); i++)
                area += triangleArea(oneRingTr[cur_v][i]);

            for (unsigned int i = 0; i < oneRing[cur_v].size(); i++)
            {
                unsigned int nei = oneRing[cur_v][i];
                if (!viewed[nei])
                {
                    viewed[nei] = true;
                    toLook.push(nei);
                }
            }
            _shapeContexts[v].addValue(dist/diskRadius, computeSCangle(cur_v, n_ref), area/3.f);
        }
    }
    _shapeContexts[v].normalize();

    if (v == refVertex)
        refSC = _shapeContexts[v];
    _sc_computed[v] = true;

    // distance between the v descriptor and the reference one
    similarity_dist[v] = refSC.distanceChi2(_shapeContexts[v]);
    switch (histoDist){
    case L2:
        similarity_dist[v] = refSC.distanceL2(_shapeContexts[v]);
        break;
    case L1:
        similarity_dist[v] = refSC.distanceL1(_shapeContexts[v]);
        break;
    case CHI2:
        similarity_dist[v] = refSC.distanceChi2(_shapeContexts[v]);
        break;
    case CHI2N:
        similarity_dist[v] = refSC.distanceChi2Normalized(_shapeContexts[v]);
        break;
    default:
        break;
    }
}

//***************************************************
//                  SVM STUFFS
//***************************************************

void PatchSelection::writeSVM(const QString & file_name, unsigned int v_indice)
{
    Trace(3, "PatchSelection::writeSVM");
    std::string file_n = qPrintable(file_name);
    ofstream out (file_n.c_str ());

    if (!out)
        exit (EXIT_FAILURE);

    vector<int> _labels;
    std::vector<std::vector<double> > _vectors;
    std::vector<unsigned int> indices_v;

    getValuesforSVM_PCAFrame(v_indice, _labels, _vectors, indices_v);
    for (unsigned int i = 0; i < indices_v.size(); i++)
    {
        out << _labels[i] ;
        for  (unsigned int j = 0; j < _vectors[i].size();j++)
            out<< " " <<j+1<<" "<<_vectors[i][j];
        out << "\n";
    }
    out.close();
}

void PatchSelection::trainSVM(){
    Trace(3, "PatchSelection::trainSVM");
    use_PK = false; // set to true to use a precomputed kernel
    vector<int> _labels;
    std::vector<std::vector<double> > _vectors;
    std::vector<unsigned int> indices_v;

    getValuesforSVM_PCAFrame(refVertex, _labels, _vectors, indices_v);
    if (!use_PK)
        _svm.setSVMparameters(C_SVC, RBF, 3, 0.1, 0.1);
    else
        _svm.setSVMparameters(C_SVC, PRECOMPUTED, 3, 0.1, 0.1);
    _svm.setCparameter(10000000.00);
    vector<int> wl;
    wl.push_back(-1);
    vector<double> _weights;
    _weights.push_back(1000.0);

    // We set gamma to the average of the average distance between each point and it closest k neighbors
    unsigned int k_neighbors =4;
    vector< double > distances_i;
    for (unsigned int i = 0; i < _vectors.size()-1; i++)
        distances_i.push_back(0.f);

    vector< vector<double> > distances;
    for (unsigned int i = 0; i < _vectors.size(); i++)
            distances.push_back(distances_i);

    for (unsigned int i = 0; i < _vectors.size()-1; i++)
    {
        for (unsigned int j = i+1; j < _vectors.size(); j++)
        {
            for (unsigned int k = 0; k < _vectors[i].size(); k++)
                distances[i][j-1] += (_vectors[i][k]-_vectors[j][k])*(_vectors[i][k]-_vectors[j][k]);
            distances[j][i] = distances[i][j-1];
        }
    }

    if (use_PK)
    {
        std::vector<std::vector<double> > _precomputed_kernel;
        for (unsigned int i = 0; i < _vectors.size(); i++)
        {
            vector< double > pk_i;
            pk_i.push_back(i+1);
            for (unsigned int j = 0; j < _vectors.size() ; j++)
            {
                pk_i.push_back( sqrt((_vectors[i][0]-_vectors[j][0])*(_vectors[i][0]-_vectors[j][0])+
                                     (_vectors[i][1]-_vectors[j][1])*(_vectors[i][1]-_vectors[j][1])+
                                     (_vectors[i][2]-_vectors[j][2])*(_vectors[i][2]-_vectors[j][2])));
            }
            _precomputed_kernel.push_back(pk_i);
        }
        _svm.setSVMproblem(_labels, _precomputed_kernel);
    } else {
        double gamma = 0.f;
        for (unsigned int i = 0; i < _vectors.size(); i++)
        {
            sort(distances[i].begin(), distances[i].end());
            double avg_i = 0.f;
            for (unsigned int j = 0; j < k_neighbors; j++)
                avg_i+= distances[i][j];
            gamma += (avg_i/k_neighbors);
        }
        gamma /= _vectors.size();
        Trace (3, " ------------------------- GAMMA " << gamma << "  Neighbors " << k_neighbors);
        _svm.setKernelGamma(gamma);
        _svm.setSVMproblem(_labels, _vectors);
    }
    _svm.setProbability(1); // 0 = doesn't give the probability associated to each label

    _svm.train();
}


void PatchSelection::predictSVM(unsigned int v_ind){
    Trace(3, "PatchSelection::trainSVM");

    vector<int> _labels;
    std::vector<std::vector<double> > _vectors;
    std::vector<std::vector<double> > _probabilities;

    // get the vectors in feature space
    std::vector<unsigned int> indices_v;
    getValuesforSVM_PCAFrame(v_ind, _labels, _vectors, indices_v);

    if (v_ind != refVertex)
        ind_vertex = v_ind;

    //    SVM prediction
    if (!use_PK)
        _svm.predictWithProb(_labels, _vectors, _probabilities);
    else {
        std::vector<std::vector<double> > _precomputed_kernel;
        for (unsigned int i = 0; i < _vectors.size(); i++)
        {
            vector< double > pk_i;
            pk_i.push_back(i+1);
            for (unsigned int j = 0; j < _vectors.size() ; j++)
            {
                pk_i.push_back( sqrt((_vectors[i][0]-_vectors[j][0])*(_vectors[i][0]-_vectors[j][0])+
                        (_vectors[i][1]-_vectors[j][1])*(_vectors[i][1]-_vectors[j][1])));
            }
            _precomputed_kernel.push_back(pk_i);
        }
        _svm.predictWithProb(_labels, _precomputed_kernel, _probabilities);
    }

    // set the vertices as selected or not depending on the SVM result
    vector<float>probaV;
    for (unsigned int i = 0; i < vertices.size(); i++)
        probaV.push_back(0.f);

    vector<int> selectedV;
    for (unsigned int i = 0; i< vertices.size(); i++)
        selectedV.push_back(-1);

    std::vector<std::pair<float, unsigned int> > probaAndV;
    unsigned int countNotNegProb = 0;
    for (unsigned int i = 0; i < _labels.size(); i++)
    {
        if (_labels[i] == 1)
        {
            selectedV[indices_v[i]] = 1;
            probaV[indices_v[i]] = _probabilities[i][0];

            if (SEE_SVM)
                probaAndV.push_back(make_pair( _probabilities[i][0], indices_v[i]));
            if (_probabilities[i][0] != 0.f)
                countNotNegProb++;
        } else {
            selectedV[indices_v[i]]=-4.f;
        }
    }


    if (SEE_SVM) // useful for debug
    {
        sort(probaAndV.begin(), probaAndV.end());

        vector<unsigned int> seed;
        for (unsigned int i = probaAndV.size()-1; i > probaAndV.size()/2.f; i--)
            seed.push_back(probaAndV[i].second);
        Trace(1, "seed.size " << seed.size() << " nb of proba " << countNotNegProb );
        for (unsigned int i = 0; i < selectedV.size(); i++)
            selectedV[i]=-1;
        for(unsigned int i = 0; i < seed.size(); i++)
            selectedV[seed[i]] = 1.f;

        bool triangleSelected;
        unsigned int cur_v;
        for (unsigned int i = 0; i< faceInSelectionK.size(); i++)
        {
            triangleSelected = true;
            for (unsigned int j = 0; j < 3; j++)
            {
                cur_v = triangles[i].getVertex(j);
                if (selectedV[cur_v] == -1)
                    triangleSelected = false;
            }
            if (triangleSelected)
                faceInSelectionK[i] = 1.5;
            else if (faceInSelectionK[i] != 1.f)
                faceInSelectionK[i] = -1.;
        }
    } else {
        // set the triangles as selected or not depending on the SVM
        bool triangleSelected;
        bool triangleInsideDisk;
        unsigned int cur_v;
        float probaT;
        for (unsigned int i = 0; i< faceInSelectionK.size(); i++)
        {
            triangleSelected = true;
            triangleInsideDisk = false;
            probaT=0.f;
            for (unsigned int j = 0; j < 3; j++)
            {
                cur_v = triangles[i].getVertex(j);
                if (selectedV[cur_v] == -4){
                    triangleInsideDisk = true;
                    triangleSelected = false;
                }else if (selectedV[cur_v] == -1 )
                    triangleSelected = false;
                else
                    probaT+= probaV[cur_v];
            }
            // erase to see triangle inside disk
            triangleInsideDisk = false;
            if (triangleSelected)
            {
                faceInSelectionK[i] =  2- probaT/3.f;
            } else if (triangleInsideDisk)
                faceInSelectionK[i]=-4.f;
            else if (faceInSelectionK[i] != 1.f  && faceInSelectionK[i] != -4.f)
                faceInSelectionK[i] = -1.;
        }
    }
}

void PatchSelection::getValuesforSVM_PCAFrame(unsigned int v_indice,
                                              std::vector< int > &_labels,
                                              std::vector<std::vector<double> > &_vectors,
                                              std::vector<unsigned int> &vertices_ind )
{
    Trace(3, "PatchSelection::getValuesforSVM_PCAFrame");

    //clear everything
    _labels.clear();
    for (unsigned int i = 0; i < _vectors.size(); i++)
        _vectors[i].clear();
    _vectors.clear();

    // find all the vertices in the set
    findVerticesInsideDisk(vertices_ind, v_indice);
    std::vector<Vec3Df> eVec;
    unsigned int flipAxis;

    // find the frame of the local set
    // using the pca and the moment difference
    findLocalFrame(vertices_ind, v_indice, eVec, flipAxis );

    // save the frame
    if (v_indice == refVertex)
    {
        for (unsigned int i = 0; i < 3; i++)
            PCAeigenVecRef[i] = eVec[i];
    } else {
        for (unsigned int i = 0; i < 3; i++)
            PCAeigenVec[i] = eVec[i];
    }

    unsigned int cur_v;
    Vec3Df c = vertices[v_indice].getPos();

    for (unsigned int i = 0; i < vertices_ind.size(); i++)
    {
        cur_v = vertices_ind[i];
        _labels.push_back(verticeInSelection[cur_v]);

        // for each vertex of the set we compute the caracteristics vector
        std::vector<double> caracteristics;
        Vec3Df p = vertices[cur_v].getPos()-c;
        // dist
        float dist = p.getLength();
        caracteristics.push_back( dist/(diskRadius) );
        // height
        float h =  Vec3Df::dotProduct(p, eVec[0]);
        caracteristics.push_back(h/(diskRadius));
        // angle
        if (dist != 0)
        {
            float alpha = Vec3Df::dotProduct(eVec[2], p/dist);
            caracteristics.push_back(alpha);
        } else
            caracteristics.push_back(0.f);
        _vectors.push_back(caracteristics);
    }
}

void PatchSelection::getValuesforSVM_withICP(std::vector< int > &_labels,
                                             std::vector<std::vector<double> > &_vectors,
                                             std::vector<unsigned int> &vertices_ind,
                                             std::vector<Vec3Df> eVec,
                                             Vec3Df center)
{
    unsigned int cur_v;

    for (unsigned int i = 0; i < vertices_ind.size(); i++)
    {
        cur_v = vertices_ind[i];
        _labels.push_back(verticeInSelection[cur_v]);

        std::vector<double> caracteristics;
        Vec3Df p = vertices[cur_v].getPos()-center;
        float u =  Vec3Df::dotProduct(p, eVec[0]);
        float v =  Vec3Df::dotProduct(p, eVec[1]);
        float w =  Vec3Df::dotProduct(p, eVec[2]);
        caracteristics.push_back(u/(diskRadius));
        caracteristics.push_back(v/(diskRadius));
        caracteristics.push_back(w/(diskRadius));
        _vectors.push_back(caracteristics);
    }
}

void PatchSelection::initICP_SVM(std::vector<unsigned int> & vertices_ind,
                                 std::vector<Vec3Df> & frame,
                                 Vec3Df & center){
    Trace(3, "PatchSelection::initICP_SVM");
    // train SVM
    std::vector< int > _labels;
    std::vector<std::vector<double> > _vectors;

    getValuesforSVM_withICP(_labels, _vectors, vertices_ind, frame, center );
    _svm.setSVMparameters(C_SVC, RBF, 3, 0.1, 0.1);
    _svm.setCparameter(1000000.00);

    vector<int> wl;
    wl.push_back(-1);
    vector<double> _weights;
    _weights.push_back(1000.0);

    QTime timer;
    timer.start();

    // We set gamma to the average of the average distance between each point and it closest k neighbors
    unsigned int k_neighbors = 4;
    vector< double > distances_i;
    for (unsigned int i = 0; i < _vectors.size()-1; i++)
        distances_i.push_back(0.f);

    double gamma = 0.f;
    unsigned int v, v_j;
    float m, d;
    for (unsigned int i = 0; i < vertices_ind.size(); i++)
    {
        v = vertices_ind[i];
        m = 0.f;
        for (unsigned int j = 0; j < oneRing[v].size(); j++)
        {
            d= 0.f;
            v_j = oneRing[v][j];
            Vec3Df p = vertices[v_j].getPos()-center;
            for (unsigned int k = 0; k < 3; k++)
                d+= (_vectors[i][k]- Vec3Df::dotProduct(p, frame[k])/diskRadius)*
                        (_vectors[i][k]- Vec3Df::dotProduct(p, frame[k])/diskRadius);
            m +=d;
        }
        gamma += m/oneRing[v].size();
    }
    gamma/=vertices_ind.size();

    Trace (4, " ------------------------- GAMMA " << gamma << "  Neighbors " << k_neighbors << " computation TIME " << timer.elapsed());
    _svm.setKernelGamma(gamma);
    _svm.setSVMproblem(_labels, _vectors);
    _svm.setProbability(1); // 0 = pas de proba
    _svm.train();
}


//***************************************************
//                  ICP STUFFS
//***************************************************

void PatchSelection::computeWeightAndNormals(const std::vector<unsigned int > & v_indices,
                                             std::vector<Vec3Df> & normals,
                                             std::vector<double> & weights)
{
    unsigned int ind;
    double area;
    for (unsigned int i = 0; i < v_indices.size(); i++)
    {
        ind = v_indices[i];
        normals.push_back(vertices[ind].getSmoothNormal());
        area  = 0.f;
        const std::vector<unsigned int> & faceV = vertices[ind].getFaces();
        for (unsigned int j = 0; j < faceV.size(); j++)
            area += triangleArea(faceV[j])/3.f ;
        weights.push_back(area);
    }

    // we normalized the weights
    double sum = 0.;
    for(unsigned int i = 0; i < weights.size(); i++)
        sum+= weights[i];

    if (sum != 1.)
    {
        for (unsigned int i = 0; i < weights.size(); i++)
            weights[i]/=sum;
    }
}


void PatchSelection::findLocalFrame( std::vector<unsigned int> & vertices_ind, unsigned int v_indice,
                                     std::vector<Vec3Df> & frame, unsigned int & flipAxis)
{
    // PCA
    float eVal[3];
    Vec3Df eVec[3];
    vector<Vec3Df> normals;
    vector<double> weights;

    // we compute a weighted pca on the normals of the vertices of the set
    // the weight of each vertex depends on its Voronoi area
    computeWeightAndNormals(vertices_ind, normals, weights);
    Vec3Df avg_norm = WeightedPCA(normals, weights, eVal, eVec);

    // we compute the frame resulting from the PCA
    // be sure all the vectors are of unit size and orthogonals
    avg_norm.normalize();
    eVec[1].normalize();
    eVec[2].normalize();
    eVec[0]=avg_norm;
    eVec[1]=Vec3Df::crossProduct(eVec[0], eVec[2]);
    eVec[2]=Vec3Df::crossProduct(eVec[0], eVec[1]);

    Vec3Df c = vertices[v_indice].getPos();

    // DIFFERENCE OF THE MOMENTS
    float moment1u = 0.f;
    float moment2u = 0.f;
    float moment1v = 0.f;
    float moment2v = 0.f;
    for (unsigned int i = 0; i < vertices_ind.size(); i++)
    {
        // compute area of the vertex
        float area = 0.f;
        for (unsigned int t = 0; t < oneRingTr[vertices_ind[i]].size(); t++)
            area += triangleArea(oneRingTr[vertices_ind[i]][t]);
        area/=3.f;

        Vec3Df p = vertices[vertices_ind[i]].getPos()-c;
        moment1u += area*(Vec3Df::dotProduct(p,eVec[2]))*(Vec3Df::dotProduct(p,eVec[0]));
        moment2u += area*(Vec3Df::dotProduct(p,eVec[2]))*(Vec3Df::dotProduct(p,eVec[2]))*
                (Vec3Df::dotProduct(p,eVec[0]));

        moment1v += area*(Vec3Df::dotProduct(p,eVec[1]))*(Vec3Df::dotProduct(p,eVec[0]));
        moment2v += area*(Vec3Df::dotProduct(p,eVec[1]))*(Vec3Df::dotProduct(p,eVec[1]))*
                (Vec3Df::dotProduct(p,eVec[0]));
    }
    moment1u/=(diskRadius);
    moment2u/=(diskRadius)*(diskRadius);
    moment1v/=(diskRadius);
    moment2v/=((diskRadius)*(diskRadius));

    if (v_indice == refVertex)
    {
        momentsRef[0]=moment1u;
        momentsRef[1]=moment2u;
        momentsRef[2]=moment1v;
        momentsRef[3]=moment2v;
    } else {
        float dist[4];
        dist[0] = std::sqrt( (moment1u-momentsRef[0])*(moment1u-momentsRef[0])+
                (moment2u-momentsRef[1])*(moment2u-momentsRef[1]));
        dist[1] = std::sqrt( (-moment1u-momentsRef[0])*(-moment1u-momentsRef[0])+
                (moment2u-momentsRef[1])*(moment2u-momentsRef[1]));
        dist[2] = std::sqrt( (moment1v-momentsRef[2])*(moment1v-momentsRef[2])
                +(moment2v-momentsRef[1])*(moment2v-momentsRef[3]));
        dist[3] = std::sqrt( (-moment1v-momentsRef[2])*(-moment1v-momentsRef[2])+
                (moment2v-momentsRef[1])*(moment2v-momentsRef[3]));

        bool flip0 = false;
        bool flip1 = false;
        if (dist[1] < dist[0])
        {
            Trace(3, "change blue vect");
            eVec[2]=-eVec[2];
            flip0 = true;
        }
        if (dist[3] < dist[2])
        {
            Trace(3, "change green vect");
            eVec[1]=-eVec[1];
            flip1 = true;
        }

        // flipAxis = 0 if no flip
        // flipAxis = 1 if eVec[1] flip
        // flipAxis = 2 if eVec[2] flip
        if ( (flip0 && !flip1))
            flipAxis = 2;
        else if ( (!flip0) && flip1)
            flipAxis = 1;
        else if ( flip0 && flip1)
            flipAxis = 3;
        else
            flipAxis = 0;
    }
    for (unsigned int i = 0; i < 3; i++)
        frame.push_back(eVec[i]);

}

void PatchSelection::initICPref(std::vector<unsigned int> & vertices_ind,
                std::vector<Vec3Df> & frame,
                Vec3Df & center)
{
    findVerticesInsideDisk(vertices_ind, refVertex);
    std::vector<Vec3Df> eVec;
    unsigned int flipAxis;
    findLocalFrame(vertices_ind, refVertex, eVec, flipAxis );

    for (unsigned int i = 0; i < 3; i++)
        PCAeigenVecRef[i] = eVec[i];
    frame.clear();
    frame.push_back(eVec[0]);
    frame.push_back(eVec[1]);
    frame.push_back(eVec[2]);
    center = vertices[refVertex].getPos();
}


void PatchSelection::computeICP(unsigned int v_indice,
                                const std::vector<unsigned int> & vertices_ind_ref,
                                std::vector<unsigned int> & vertices_ind,
                                Matrix & firstTransform, Matrix & icpTransform,
                                std::vector<Vec3Df> & loc_frame_icp,
                                Vec3Df & centerICP)
{
    Trace(3, "PatchSelection::computeICP");
    QTime icp_time;
    if (SPY_TIME)
    {
        icp_time.start();
    }
    findVerticesInsideDisk(vertices_ind, v_indice);

    std::vector<Vec3Df> eVec;
    unsigned int flipAxis;
    findLocalFrame(vertices_ind, v_indice, eVec, flipAxis );
    if (flipAxis == 3)
        flipAxis =1;

    ind_vertex = v_indice;
    for (unsigned int i = 0; i < 3; i++)
        PCAeigenVec[i] = eVec[i];

    // to express the points in the ref frame
    // center of the reference patch
    Matrix centerRef(4, 1);
    for (unsigned int i = 0; i < 3; i++)
        centerRef.val[i][0] = (vertices[refVertex].getPos())[i];
    centerRef.val[3][0] = 1.f;

    // reference rotation and translation
    Matrix Rref = Matrix::eye(4);
    Matrix Tref = Matrix::eye(4);
    for (unsigned int i = 0; i < 3; i ++)
    {
        for (unsigned int j = 0; j < 3; j++)
        {
            Rref.val[i][j]= PCAeigenVecRef[i][j];
        }
        Tref.val[i][3] = -centerRef.val[i][0];
    }

    // add a symmetry transformation if needed
    Matrix Symmetry = Matrix::eye(4);
    if (flipAxis == 1)
        Symmetry.val[1][1]= -1.f;
    else if (flipAxis == 2)
        Symmetry.val[2][2] = -1.f;

    // local frame with axis flipped if needed
    Matrix uLoc(4, 1);
    Matrix vLoc(4, 1);
    Matrix wLoc(4 ,1);
    for (unsigned int i = 0; i < 3; i++)
    {
        uLoc.val[i][0] = PCAeigenVec[0][i];
        if (flipAxis == 1)
            vLoc.val[i][0] = -PCAeigenVec[1][i];
        else
            vLoc.val[i][0] = PCAeigenVec[1][i];
        if (flipAxis == 2)
            wLoc.val[i][0] = -PCAeigenVec[2][i];
        else
            wLoc.val[i][0] = PCAeigenVec[2][i];
    }
    uLoc.val[3][0] = 0.0;
    vLoc.val[3][0] = 0.0;
    wLoc.val[3][0] = 0.0;

    Matrix Rotation = Matrix::eye(4);
    for (unsigned int i = 0; i < 3; i++)
    {
        Rotation.val[i][0] = uLoc.val[i][0];
        Rotation.val[i][1] = vLoc.val[i][0];
        Rotation.val[i][2] = wLoc.val[i][0];
    }
    Matrix Tloc = Matrix::eye(4);
    for (unsigned int i = 0; i< 3; i++)
        Tloc.val[i][3] = (vertices[v_indice].getPos())[i];

    ICPpointsRef.clear();
    // initial transformation between the two frames
    firstTransform =  Tloc*Rotation* Symmetry* Rref*Tref;

    // we express all the points of the reference set
    // in the local frame
    for (unsigned int i = 0; i < vertices_ind_ref.size(); i++)
    {
        if (verticeInSelection[vertices_ind_ref[i]] == 1.0)
        {
            Matrix p(4,1);
            Vec3Df p_v =  vertices[vertices_ind_ref[i]].getPos();
            for (unsigned int j = 0; j < 3; j++)
                p.val[j][0] = p_v[j];
            p.val[3][0] = 1.0;
            Matrix p_in_local_f(4,1);
            p_in_local_f = firstTransform* p;
            ICPpointsRef.push_back(Vec3Df( p_in_local_f.val[0][0],
                                           p_in_local_f.val[1][0],
                                           p_in_local_f.val[2][0]));
        }
    }

    // center of reference transformed to be on the local frame
    Matrix centerRefT(4,1);
    centerRefT=  Tloc*Rotation*Symmetry* Rref*Tref*centerRef;

    // set the cloud points for the ICP
    double* T = (double*) calloc(3*ICPpointsRef.size(), sizeof(double));
    for (unsigned int i = 0; i < ICPpointsRef.size(); i++)
        for (unsigned int j = 0; j < 3; j++)
            T[i*3+j] = (double) ICPpointsRef[i][j];

    Vec3Df p;
    double* M = (double*) calloc(3*vertices_ind.size(), sizeof(double));
    for (unsigned int i = 0; i<vertices_ind.size(); i++)
    {
        p = vertices[vertices_ind[i]].getPos();
        M[i*3+0]= (double) p[0];
        M[i*3+1]= (double) p[1];
        M[i*3+2]= (double) p[2];
    }
    // Compute the ICP
    Matrix t(3,1);
    Matrix R0 = Matrix::eye(3);
    IcpPointToPlane icp(M, vertices_ind.size(), 3);
//    icp.setMinDeltaParam(1e-8);
    icp.fit(T, ICPpointsRef.size(), R0, t, -1);

    Vec3Df tr = Vec3Df(t.val[0][0], t.val[1][0], t.val[2][0]) ;
    float tr_l = tr.getLength();
    icpTransform = Matrix::eye(4);

    if (tr_l < diskRadius*3.f){
        for (unsigned int i = 0; i < 3; i++)
        {
            for (unsigned int j = 0; j < 3; j++)
                icpTransform.val[i][j] = R0.val[i][j];
            icpTransform.val[i][3] = t.val[i][0];
        }
    }

    // compute the position of the center after the icp
    Matrix centerRefICP = icpTransform*centerRefT;
    for (unsigned int i = 0; i < 3; i++)
        centerVecDeformed[i] = centerRefICP.val[i][0];
    centerICP = centerVecDeformed;

    // compute the local frame after the icp
    for (unsigned int i = 0; i < 3; i++)
    {
        uLoc.val[i][0] = PCAeigenVec[0][i];
        vLoc.val[i][0] = PCAeigenVec[1][i];
        wLoc.val[i][0] = PCAeigenVec[2][i];
    }

    Matrix uLocICP(4, 1);
    Matrix vLocICP(4, 1);
    Matrix wLocICP(4 ,1);

    uLocICP= icpTransform*uLoc;
    vLocICP= icpTransform*vLoc;
    wLocICP= icpTransform*wLoc;

    for (unsigned int i = 0; i < 3; i++)
    {
        PCAeigenVecDeformed[0][i] = uLocICP.val[i][0];
        PCAeigenVecDeformed[1][i] = vLocICP.val[i][0];
        PCAeigenVecDeformed[2][i] = wLocICP.val[i][0];
    }

    // compute the reference vertices positions after ICP
    // necessary only for visualization
    ICPpointsRefTransformed.clear();
    for (unsigned int i = 0; i < ICPpointsRef.size(); i++)
    {
        Matrix p(4,1);
        Matrix pICP(4,1);
        for (unsigned int j = 0; j < 3; j++)
            p.val[j][0] = ICPpointsRef[i][j];
        p.val[3][0] = 1;
        pICP = icpTransform*p;
        ICPpointsRefTransformed.push_back(Vec3Df(pICP.val[0][0], pICP.val[1][0], pICP.val[2][0]));
    }

    free(M);
    free(T);

    loc_frame_icp.push_back(PCAeigenVecDeformed[0]);
    loc_frame_icp.push_back(PCAeigenVecDeformed[1]);
    loc_frame_icp.push_back(PCAeigenVecDeformed[2]);

    if (SPY_TIME){
        timing_sim t;
        t.time =  icp_time.elapsed();
        t.sel_nb = nb_of_global_sel;
        t.sim_nb = simSelectionsNb;
        timing_ICP.push_back(t);
    }
}

void PatchSelection::computeBasicAlgoFromICP(Matrix & firstTransform, Matrix & icpTransform ){
    Trace(3, " computeBasicAlgoFromICP ");

    // compute the transformation of the stroke points
    std::vector<Vec3Df> transformedStrokePoints;
    for (unsigned int i = 0; i < strokeForSimiliarity.size(); i++)
    {
        Matrix p(4,1);
        for (unsigned int j = 0; j < 3; j++)
            p.val[j][0] = (strokeForSimiliarity[i].p)[j];
        p.val[3][0] = 1.f;
        Matrix pTransformed(4,1);
        pTransformed = icpTransform*firstTransform*p;
        transformedStrokePoints.push_back(Vec3Df(pTransformed.val[0][0], pTransformed.val[1][0], pTransformed.val[2][0]));
    }

    std::vector< UserStroke::strokePoints > stroke;
    for (unsigned int i = 0; i < strokeForSimiliarity.size(); i++)
    {
        stroke.push_back(strokeForSimiliarity[i]);
        stroke[i].p = transformedStrokePoints[i];
    }

    // run the selection algorithm for these virtual strokes
    vector< int > viewedFaces;
    for (unsigned int i = 0; i < triangles.size(); i++)
        viewedFaces.push_back(-1);

    vector< unsigned int > unsortedFaces;

    initializeElementsToClassify(0, stroke, viewedFaces, unsortedFaces);

    classifyUnsortedFaces(viewedFaces, unsortedFaces);
    for (unsigned int i = 0; i < viewedFaces.size(); i++)
    {
        if (viewedFaces[i] == 2)
                faceInSelectionK[i] = 1.1f;
    }
    Trace(1, " computeBasicAlgoFromICP end ");
}



