/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "SelectionsManager.h"

void SelectionsManager::init(const Mesh &m)
{
    Trace(2, "SelectionsManager::init ");
    _comp->clear();
    _part->clear();
    _patch->clear();

    _comp->init(m);
    _part->init(m);

    mesh = m;
    CotanLaplacian* l = _part->getCotanLaplacian();
    _patch->setMeshLaplacian(l);

    _patch->init(m);
    _curSelection = _patch;

//    connect(_curSelection, SIGNAL(similarityMapComputed(int)), this, SLOT(similarityMapComputed(int)));
    _currentSelections.clear();
    _selections.clear();
    for (unsigned int i = 0; i < m.getTriangles().size(); i++)
    {
        _currentSelections.push_back(-1.f);
        _selections.push_back(-1.f);
    }
    nnsSelectionsNb = 0;
    allSelectionsNb = 0;
}


void SelectionsManager::startSelection(SelectionWidget* _widget )
{
    Trace(2, "SelectionsManager::startSelection ");
    UserStroke & currentStroke = _widget->getStroke();
    UserStroke & refineStroke = _widget->getRefineStroke();


    Trace (3, " nb of strokes " << currentStroke.getNumberOfStrokes());
    switch(_widget->getMode())
    {
    case SelectionWidget::DRAW:
        if (currentStroke.getNumberOfStrokes() == 1)
        {
            if (_type == COMP)
                _curSelection = _comp ;
            else if (_type == PART)
                _curSelection = _part ;
            else if (_type == PATCH)
                _curSelection = _patch ;
            _curSelection->startSelection(currentStroke);
        } else if (currentStroke.getNumberOfStrokes() > 1) {
            _curSelection->updateSelection(currentStroke);
        }
        break;
    case SelectionWidget::DEL:
        deleteSelectionUnderStroke(currentStroke);
        break;
    case SelectionWidget::ERASE:
//        if (currentStroke.getNumberOfStrokes() >= 1)
//        {
            if (_type == COMP)
                _curSelection = _comp ;
            else if (_type == PART)
                _curSelection = _part ;
            else if (_type == PATCH)
                _curSelection = _patch ;
            _curSelection->eraseSelection(currentStroke);

//        } else if (currentStroke.getNumberOfStrokes() == 0)
//        {
//            clearCurrentSelections();
//            _widget->clear();
//        }
        break;
    case SelectionWidget::REFINE:
        _curSelection->refineSelection(refineStroke);
        break;
    default:
        break;
    }
    _currentSelections = _curSelection->getSelections();
}

void SelectionsManager::updateSelection(SelectionWidget* _widget)
{
    Trace(4, " update selection ");
    UserStroke & currentStroke = _widget->getStroke();
    UserStroke & refineStroke = _widget->getRefineStroke();
    switch(_widget->getMode())
    {
    case SelectionWidget::DRAW:
        _curSelection->updateSelection(currentStroke);
        break;
    case SelectionWidget::ERASE:
    case SelectionWidget::DEL:
        _curSelection->eraseSelection(currentStroke);
        break;
    case SelectionWidget::REFINE:
        _curSelection->refineSelection(refineStroke);
        break;
    default:
        break;
    }
    _currentSelections = _curSelection->getSelections();
}

void SelectionsManager::deleteSelectionUnderStroke(UserStroke &stroke){
    std::vector<UserStroke::strokePoints> strokePoints = stroke.getBGPoints();
    assert (!strokePoints.empty());

    double d0, d1, d2;
    unsigned int triangle_under_point;
    mesh.barycentricCoordinates(strokePoints[strokePoints.size()-1].p,
                                triangle_under_point, d0, d1, d2);

    const std::vector<Triangle> & T = mesh.getTriangles ();
    Trace(1, " delete triangle " << triangle_under_point);
    float value_to_delete = _selections[triangle_under_point];
    if (value_to_delete != -1.f){
        _selections[triangle_under_point] = -1;
        Trace(1, " triangle deleted" );

        std::queue<unsigned int> toLook;

        std::vector<bool> viewed;
        for (unsigned int i = 0; i < T.size(); i++)
            viewed.push_back(false);
        const std::vector< std::vector <unsigned int> > neighbors = mesh.getFaceNeighbors();
        viewed[triangle_under_point] = true;
        for (unsigned int i = 0; i < neighbors[triangle_under_point].size(); i++ )
        {
            toLook.push(neighbors[triangle_under_point][i]);
            viewed[neighbors[triangle_under_point][i]] = true;
        }

        unsigned int cur_f;
        int val_to_del = (int) value_to_delete;
        while (!toLook.empty())
        {
            cur_f = toLook.front();
            toLook.pop();

            if ( (int) _selections[cur_f] == val_to_del)
            {
                for (unsigned int i = 0; i < neighbors[cur_f].size(); i++)
                {
                    unsigned int nei = neighbors[cur_f][i];
                    if (!viewed[nei])
                    {
                        viewed[nei] = true;
                        toLook.push(nei);
                    }
                }
                _selections[cur_f] = -1.f;
            }
        }
    }
}

void SelectionsManager::reverseCurrentSelection(){

    for (unsigned int i = 0; i < _currentSelections.size(); i++)
    {
        if (_currentSelections[i] != -1.f)
            _currentSelections[i] = -1.f;
        else {
            if (_type == COMP)
                _currentSelections[i] = 100.f ;
            else if (_type == PART)
                _currentSelections[i] = 1000.f ;
            else
                _currentSelections[i] = 1.f;
        }
    }
    _curSelection->setSelections(_currentSelections);
}

void SelectionsManager::deleteAll()
{
    _patch->clearSelections();
    _part->clearSelections();
    _comp->clearSelections();
    _curSelection->clearSelections();
    _curSelection->getSelections();

    for (unsigned int i = 0; i < _selections.size(); i++)
        _selections[i] = -1.f;

    for (unsigned int i = 0; i < _currentSelections.size(); i++)
        _currentSelections[i] = -1.f;

}

void SelectionsManager::endSelection(SelectionWidget* _widget)
{
    UserStroke & currentStroke = _widget->getStroke();
    switch(_widget->getMode())
    {
    case SelectionWidget::DRAW:
        _curSelection->endSelection(currentStroke);
        break;
    case SelectionWidget::ERASE:
        Trace(1, " end Selection end ");
        _curSelection->eraseSelection(currentStroke);
        break;
    case SelectionWidget::DEL:
        break;
    default:
        break;
    }
    _currentSelections = _curSelection->getSelections();

}

void SelectionsManager::saveCurrentSelections() {
    recordCurrentSelections();
    clearCurrentSelections();
}

void SelectionsManager::clearCurrentSelections(){
    Trace(2, "SelectionsManager::clearCurrentSelections ");
    _curSelection->clearSelections();
    _currentSelections = _curSelection->getSelections();
}

void SelectionsManager::recordCurrentSelections(){
    Trace(2, "SelectionsManager::recordCurrentSelections ");
    for (unsigned int i = 0; i < _selections.size(); i++)
    {
        if (_currentSelections[i] != -1)
        {
            _selections[i] = nnsSelectionsNb + _currentSelections[i];
        }
    }
    nnsSelectionsNb ++;
    allSelectionsNb += _curSelection->getSimSelectionsNb();
}

void SelectionsManager::searchSimilarSelections(UserStroke & stroke){
    Trace(2, "SelectionsManager::searchSimilarSelections");
    if (_type != NONE)
    {
        _curSelection->findSimilarSelections(stroke);
        _currentSelections = _curSelection->getSelections();
    }
}

void SelectionsManager::openSVMselection(const QString &filename)
{
    if (_type == PATCH)
    {
        Trace(1, " Lalalaaa \n");
        _patch->openSVMselection(filename);
        _currentSelections = _curSelection->getSelections();
    }
}
void SelectionsManager::changeSimilarityThreshold(int val){
    Trace(1, "SelectionsManager::changeSimilarityThreshold " << val);
    _patch->setSimiliarityThreshold(val);
    _part->setSimiliarityThreshold(val);
    _comp->setSimiliarityThreshold(val);
    if (_type != NONE)
    {
        _curSelection->changeSimiliarityThreshold(val);
        _currentSelections = _curSelection->getSelections();
    }
}

