/*
 *     SimSelect - Similarity-based Selection for 3D Surfaces
 *
 *     Authors : Emilie GUY, Jean-Marc THIERY and Tamy BOUBEKEUR
 *
 *     Copyright © 2012-2014 Telecom ParisTech - CNRS LTCI - Institut Mines Telecom
 *              All rights reserved
 *
 * This file is part of the SimSelect project
 *
 * SimSelect is free software: you can redistribute it and/or modify
 *
 * it under the terms of the GNU Lesser General Public License as published by
 *
 * the Free Software Foundation, either version 3 of the License, or
 *
 * (at your option) any later version.
 *
 * SimSelect is distributed in the hope that it will be useful,
 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 *
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "MathTools.h"

#include <stdio.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

using namespace std;


// *****************************
// Principal Component Analysis.
// *****************************

inline double cov (unsigned int i, unsigned int j, 
                   const Vec3Df & a, const Vec3Df & b, 
                   const Vec3Df & mean) {
    return static_cast<double>((a[i] - mean[i]) * (b[j] - mean[j]));
}

void principalComponentAnalysis (const vector<Vec3Df> & points,
                                 float * eigenValues, Vec3Df * eigenVectors) {
    // mean computation
    Vec3Df mean;
    for (vector<Vec3Df>::const_iterator it = points.begin ();
            it != points.end ();
            it++)
        mean += *it;
    mean /= points.size ();

    // covariance matrix computation
    vector<Vec3Df> covariance;
    covariance.resize (points.size ());
    double covMat[9];

    for (unsigned int i = 0; i < 9; i++)
        covMat[i] = 0.0;
    for (vector<Vec3Df>::const_iterator it = points.begin ();
            it != points.end ();
            it++)
        for (unsigned int i = 0; i < 3; i++)
            for (unsigned int j = 0; j < 3; j++)
                covMat[i*3+j] += cov (i, j, *it, *it, mean);

    gsl_matrix_view m = gsl_matrix_view_array (covMat, 3, 3);
    gsl_vector *eval = gsl_vector_alloc (3);
    gsl_matrix *evec = gsl_matrix_alloc (3, 3);
    gsl_eigen_symmv_workspace * w = gsl_eigen_symmv_alloc (3);
    gsl_eigen_symmv (&m.matrix, eval, evec, w);
    gsl_eigen_symmv_free (w);
    gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_ABS_ASC);
    for (unsigned int i = 0; i < 3; i++) {
        eigenValues[i] = gsl_vector_get (eval, i);
        gsl_vector_view evec_i = gsl_matrix_column (evec, i);
        eigenVectors[i] = Vec3Df (gsl_vector_get (&evec_i.vector, 0),
                                  gsl_vector_get (&evec_i.vector, 1),
                                  gsl_vector_get (&evec_i.vector, 2));
    }
    gsl_vector_free(eval);
    gsl_matrix_free(evec);
}

void principalComponentAnalysisOnNormals (const vector<Vertex> & vertices,
                                 const vector<unsigned int> & indices,
                                 float * eigenValues, Vec3Df * eigenVectors) {

    // mean computation
    Vec3Df mean;
    for (vector<unsigned int>::const_iterator it = indices.begin ();
            it != indices.end ();
            it++)
        mean += vertices[(*it)].getNormal ();
    mean /= indices.size ();


    // covariance matrix computation
    double covMat[9];
    for (unsigned int i = 0; i < 9; i++)
        covMat[i] = 0.0;
    for (vector<unsigned int>::const_iterator it = indices.begin ();
            it != indices.end ();
            it++) {
        const Vec3Df & p = vertices[(*it)].getNormal ();
        for (unsigned int i = 0; i < 3; i++)
            for (unsigned int j = 0; j < 3; j++)
                covMat[i*3+j] += cov (i, j, p, p, mean);
    }

    gsl_matrix_view m = gsl_matrix_view_array (covMat, 3, 3);
    gsl_vector *eval = gsl_vector_alloc (3);
    gsl_matrix *evec = gsl_matrix_alloc (3, 3);
    gsl_eigen_symmv_workspace * w = gsl_eigen_symmv_alloc (3);
    gsl_eigen_symmv (&m.matrix, eval, evec, w);
    gsl_eigen_symmv_free (w);
    gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_ABS_ASC);
    for (unsigned int i = 0; i < 3; i++) {
        eigenValues[i] = gsl_vector_get (eval, i);
        gsl_vector_view evec_i = gsl_matrix_column (evec, i);
        eigenVectors[i] = Vec3Df (gsl_vector_get (&evec_i.vector, 0),
                                  gsl_vector_get (&evec_i.vector, 1),
                                  gsl_vector_get (&evec_i.vector, 2));
    }
    gsl_vector_free(eval);
    gsl_matrix_free(evec);
}

Vec3Df WeightedPCA(const vector<Vec3Df> & datas,
                  const vector<double> & weights,
                  float* eigenValues, Vec3Df* eigenVectors)
{
    assert(datas.size() == weights.size());

    // Assert that the Sum_Of_Weights = 1
    double w_sum = 0;
    for (unsigned int i=0; i < weights.size(); i++)
        w_sum+=weights[i];
    Trace (4, " w_sum " << w_sum);
    assert(w_sum >= 0.95 && w_sum <= 1.05);

    // compute the weighted average
    Vec3Df w_avg = Vec3Df(0., 0., 0.);
    for (unsigned int i =0; i< datas.size(); i++)
        w_avg += weights[i]*datas[i];

    // compute the weighted covariance
    double covMat[9];
    for (unsigned int i = 0; i < 9; i++)
        covMat[i] = 0.0;
    for (unsigned int i=0; i < datas.size(); i++)
    {
        const Vec3Df & p = datas[i];
        for (unsigned int j = 0; j < 3; j++)
            for (unsigned int k = 0; k < 3; k++)
                covMat[j*3+k] += cov (j, k, p, p, w_avg)*weights[i];
    }

   // compute the sum of weights*weights
   double w2_sum =0.;
   for (unsigned int i=0; i < weights.size(); i++)
       w2_sum+=(weights[i]*weights[i]);

   // weight the covMat by 1/(1-w2_sum)
   double coeff = 1./(1.-w2_sum);
   for (unsigned int i=0; i < 9; i++)
       covMat[i]/=coeff;


   gsl_matrix_view m = gsl_matrix_view_array (covMat, 3, 3);
   gsl_vector *eval = gsl_vector_alloc (3);
   gsl_matrix *evec = gsl_matrix_alloc (3, 3);
   gsl_eigen_symmv_workspace * w = gsl_eigen_symmv_alloc (3);
   gsl_eigen_symmv (&m.matrix, eval, evec, w);
   gsl_eigen_symmv_free (w);
   gsl_eigen_symmv_sort (eval, evec, GSL_EIGEN_SORT_ABS_ASC);
   for (unsigned int i = 0; i < 3; i++) {
       eigenValues[i] = gsl_vector_get (eval, i);
       gsl_vector_view evec_i = gsl_matrix_column (evec, i);
       eigenVectors[i] = Vec3Df (gsl_vector_get (&evec_i.vector, 0),
                                 gsl_vector_get (&evec_i.vector, 1),
                                 gsl_vector_get (&evec_i.vector, 2));
   }
   gsl_vector_free(eval);
   gsl_matrix_free(evec);
   return w_avg;
}

